from uuid import UUID
import pandas as pd
from typing import Optional, List
import sempy_labs._icons as icons
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _base_api,
    _create_dataframe,
    resolve_item_id,
)


def create_external_data_share(
    item_name: str,
    item_type: str,
    paths: str | List[str],
    recipient: str,
    workspace: Optional[str | UUID] = None,
):
    """
    Creates an external data share for a given path or list of paths in the specified item.

    This is a wrapper function for the following API: `External Data Shares - Create External Data Share <https://learn.microsoft.com/rest/api/fabric/core/external-data-shares/create-external-data-share>`_.

    Parameters
    ----------
    item_name : str
        The item name.
    item_type : str
        The `item type <https://learn.microsoft.com/rest/api/fabric/core/items/list-items?tabs=HTTP#itemtype>`_.
    paths : str | List[str]
        The path or list of paths that are to be externally shared. Currently, only a single path is supported.
    recipient : str
        The email address of the recipient.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_id = resolve_item_id(item=item_name, type=item_type, workspace=workspace_id)

    if isinstance(paths, str):
        paths = [paths]

    payload = {"paths": paths, "recipient": {"userPrincipalName": recipient}}

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/items/{item_id}/externalDataShares",
        method="post",
        status_codes=201,
        payload=payload,
    )
    print(
        f"{icons.green_dot} An external data share was created for the '{item_name}' {item_type} within the '{workspace_name}' workspace for the {paths} paths."
    )


def revoke_external_data_share(
    external_data_share_id: UUID,
    item_name: str,
    item_type: str,
    workspace: Optional[str | UUID] = None,
):
    """
    Revokes the specified external data share. Note: This action cannot be undone.

    This is a wrapper function for the following API: `External Data Shares - Revoke External Data Share <https://learn.microsoft.com/rest/api/fabric/core/external-data-shares/revoke-external-data-share`_.

    Parameters
    ----------
    external_data_share_id : uuid.UUID
        The external data share ID.
    item_name : str
        The item name.
    item_type : str
        The `item type <https://learn.microsoft.com/rest/api/fabric/core/items/list-items?tabs=HTTP#itemtype>`_.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_id = resolve_item_id(item=item_name, type=item_type, workspace=workspace_id)

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/items/{item_id}/externalDataShares/{external_data_share_id}/revoke",
        method="post",
    )
    print(
        f"{icons.green_dot} The '{external_data_share_id}' external data share for the '{item_name}' {item_type} within the '{workspace_name}' workspace has been revoked."
    )


def list_external_data_shares_in_item(
    item_name: str, item_type: str, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Returns a list of the external data shares that exist for the specified item.

    This is a wrapper function for the following API: `External Data Shares - List External Data Shares In Item <https://learn.microsoft.com/rest/api/fabric/core/external-data-shares/list-external-data-shares-in-item`_.

    Parameters
    ----------
    item_name : str
        The item name.
    item_type : str
        The `item type <https://learn.microsoft.com/rest/api/fabric/core/items/list-items?tabs=HTTP#itemtype>`_.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of the external data shares that exist for the specified item.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_id = resolve_item_id(item=item_name, type=item_type, workspace=workspace_id)

    columns = {
        "External Data Share Id": "string",
        "Paths": "string",
        "Creator Principal Id": "string",
        "Creator Principal Type": "string",
        "Recipient User Principal Name": "string",
        "Status": "string",
        "Expiration Time UTC": "string",
        "Workspace Id": "string",
        "Item Id": "string",
        "Item Name": "string",
        "Item Type": "string",
        "Invitation URL": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/items/{item_id}/externalDataShares",
        uses_pagination=True,
    )
    dfs = []

    for r in responses:
        for i in r.get("value", []):
            item_id = i.get("itemId")
            new_data = {
                "External Data Share Id": i.get("id"),
                "Paths": [i.get("paths")],
                "Creator Principal Id": i.get("creatorPrincipal", {}).get("id"),
                "Creator Principal Type": i.get("creatorPrincipal", {}).get("type"),
                "Recipient User Principal Name": i.get("recipient", {}).get(
                    "userPrincipalName"
                ),
                "Status": i.get("status"),
                "Expiration Time UTC": i.get("expriationTimeUtc"),
                "Workspace Id": i.get("workspaceId"),
                "Item Id": item_id,
                "Item Name": item_name,
                "Item Type": item_type,
                "Invitation URL": i.get("invitationUrl"),
            }
            dfs.append(pd.DataFrame(new_data, index=[0]))
    df = pd.concat(dfs, ignore_index=True)

    return df



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    _is_valid_uuid,
    resolve_workspace_name_and_id,
    _update_dataframe_datatypes,
    _base_api,
    _create_dataframe,
    resolve_item_id,
)
from uuid import UUID
import sempy_labs._icons as icons
from sempy_labs._gateways import _resolve_gateway_id


def delete_connection(connection: str | UUID):
    """
    Delete a connection.

    This is a wrapper function for the following API: `Connections - Delete Connection <https://learn.microsoft.com/rest/api/fabric/core/connections/delete-connection>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    connection : str | uuid.UUID
        The connection name or ID.
    """

    connection_id = _resolve_connection_id(connection)
    _base_api(
        request=f"/v1/connections/{connection_id}", client="fabric_sp", method="delete"
    )
    print(f"{icons.green_dot} The '{connection}' connection has been deleted.")


def delete_connection_role_assignment(connection: str | UUID, role_assignment_id: UUID):
    """
    Delete the specified role assignment for the connection.

    This is a wrapper function for the following API: `Connections - Delete Connection Role Assignment <https://learn.microsoft.com/rest/api/fabric/core/connections/delete-connection-role-assignment>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    connection : str | uuid.UUID
        The connection name or ID.
    role_assignment_id : uuid.UUID
        The role assignment ID.
    """

    connection_id = _resolve_connection_id(connection)
    _base_api(
        request=f"/v1/connections/{connection_id}/roleAssignments/{role_assignment_id}",
        client="fabric_sp",
        method="delete",
    )

    print(
        f"{icons.green_dot} The '{role_assignment_id}' role assignment Id has been deleted from the '{connection}' connection."
    )


def _resolve_connection_id(connection: str | UUID) -> UUID:

    dfC = list_connections()
    if _is_valid_uuid(connection):
        dfC_filt = dfC[dfC["Connection Id"] == connection]
    else:
        dfC_filt = dfC[dfC["Connection Name"] == connection]

    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{connection}' is not a valid connection."
        )

    return dfC_filt["Connection Id"].iloc[0]


def list_connection_role_assignments(connection: str | UUID) -> pd.DataFrame:
    """
    Returns a list of connection role assignments.

    This is a wrapper function for the following API: `Connections - List Connection Role Assignments <https://learn.microsoft.com/rest/api/fabric/core/connections/list-connection-role-assignments>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    connection : str | uuid.UUID
        The connection name or ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of connection role assignments.
    """

    connection_id = _resolve_connection_id(connection)

    columns = {
        "Connection Role Assignment Id": "string",
        "Principal Id": "string",
        "Principal Type": "string",
        "Role": "string",
    }

    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request=f"/v1/connections/{connection_id}/roleAssignments",
        client="fabric_sp",
        uses_pagination=True,
    )

    for r in responses:
        for v in r.get("value", []):
            new_data = {
                "Connection Role Assignment Id": v.get("id"),
                "Principal Id": v.get("principal", {}).get("id"),
                "Principal Type": v.get("principal", {}).get("type"),
                "Role": v.get("role"),
            }

            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def list_connections() -> pd.DataFrame:
    """
    Lists all available connections.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing all available connections.
    """

    columns = {
        "Connection Id": "string",
        "Connection Name": "string",
        "Gateway Id": "string",
        "Connectivity Type": "string",
        "Connection Path": "string",
        "Connection Type": "string",
        "Privacy Level": "string",
        "Credential Type": "string",
        "Single Sign On Type": "string",
        "Connection Encryption": "string",
        "Skip Test Connection": "bool",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request="/v1/connections", client="fabric_sp", uses_pagination=True
    )

    for r in responses:
        for i in r.get("value", []):
            connection_details = i.get("connectionDetails", {})
            credential_details = i.get("credentialDetails", {})

            new_data = {
                "Connection Id": i.get("id"),
                "Connection Name": i.get("displayName"),
                "Gateway Id": i.get("gatewayId"),
                "Connectivity Type": i.get("connectivityType"),
                "Connection Path": connection_details.get("path"),
                "Connection Type": connection_details.get("type"),
                "Privacy Level": i.get("privacyLevel"),
                "Credential Type": (
                    credential_details.get("credentialType")
                    if credential_details
                    else None
                ),
                "Single Sign On Type": (
                    credential_details.get("singleSignOnType")
                    if credential_details
                    else None
                ),
                "Connection Encryption": (
                    credential_details.get("connectionEncryption")
                    if credential_details
                    else None
                ),
                "Skip Test Connection": (
                    credential_details.get("skipTestConnection")
                    if credential_details
                    else None
                ),
            }

            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


def list_item_connections(
    item_name: str, item_type: str, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows the list of connections that the specified item is connected to.

    This is a wrapper function for the following API: `Items - List Item Connections <https://learn.microsoft.com/rest/api/fabric/core/items/list-item-connections>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    item_name : str
        The item name.
    item_type : str
        The `item type <https://learn.microsoft.com/rest/api/fabric/core/items/update-item?tabs=HTTP#itemtype>`_.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the list of connections that the specified item is connected to.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_type = item_type[0].upper() + item_type[1:]
    item_id = resolve_item_id(item=item_name, type=item_type, workspace=workspace_id)

    columns = {
        "Connection Name": "string",
        "Connection Id": "string",
        "Connectivity Type": "string",
        "Connection Type": "string",
        "Connection Path": "string",
        "Gateway Id": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/items/{item_id}/connections",
        client="fabric_sp",
        uses_pagination=True,
    )

    for r in responses:
        for v in r.get("value", []):
            new_data = {
                "Connection Name": v.get("displayName"),
                "Connection Id": v.get("id"),
                "Connectivity Type": v.get("connectivityType"),
                "Connection Type": v.get("connectionDetails", {}).get("type"),
                "Connection Path": v.get("connectionDetails", {}).get("path"),
                "Gateway Id": v.get("gatewayId"),
            }

            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def _list_supported_connection_types(
    gateway: Optional[str | UUID] = None, show_all_creation_methods: bool = False
) -> pd.DataFrame:

    url = f"/v1/connections/supportedConnectionTypes?showAllCreationMethods={show_all_creation_methods}&"
    if gateway is not None:
        gateway_id = _resolve_gateway_id(gateway)
        url += f"gatewayId={gateway_id}"

    columns = {
        "Connection Type": "string",
        "Creation Method": "string",
        "Supported Credential Types": "string",
        "Supported Connection Encryption Types": "string",
        "Supports Skip Test Connection": "bool",
    }
    df = _create_dataframe(columns=columns)

    url = url.rstrip("&")
    responses = _base_api(request=url, client="fabric_sp", uses_pagination=True)

    records = []
    for r in responses:
        for v in r.get("value", []):
            records.append(
                {
                    "Connection Type": v.get("type"),
                    "Creation Method": v["creationMethods"][0]["name"],
                    "Supported Credential Types": v.get("supportedCredentialTypes"),
                    "Supported Connection Encryption Types": v.get(
                        "supportedConnectionEncryptionTypes"
                    ),
                    "Supports Skip Test Connection": v.get(
                        "supportsSkipTestConnection"
                    ),
                }
            )

    if records:
        df = pd.DataFrame(records)

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


def create_cloud_connection(
    name: str,
    server_name: str,
    database_name: str,
    user_name: str,
    password: str,
    privacy_level: str,
    connection_encryption: str = "NotEncrypted",
    skip_test_connection: bool = False,
):
    """
    Creates a shared cloud connection.

    This is a wrapper function for the following API: `Connections - Create Connection <https://learn.microsoft.com/rest/api/fabric/core/connections/create-connection>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name : str
        The name of the connection.
    server_name : str
        The name of the server.
    database_name : str
        The name of the database.
    user_name : str
        The username.
    password : str
        The password.
    privacy_level : str
        The `privacy level <https://learn.microsoft.com/rest/api/fabric/core/connections/create-connection?tabs=HTTP#privacylevel>`_ of the connection.
    connection_encryption : str, default="NotEncrypted"
        The connection encrpytion.
    skip_test_connection: bool, default=False
        If True, skips the test connection.
    """

    payload = {
        "connectivityType": "ShareableCloud",
        "displayName": name,
        "connectionDetails": {
            "type": "SQL",
            "creationMethod": "SQL",
            "parameters": [
                {
                    "dataType": "Text",
                    "name": "server",
                    "value": server_name,
                },
                {
                    "dataType": "Text",
                    "name": "database",
                    "value": database_name,
                },
            ],
        },
        "privacyLevel": privacy_level,
        "credentialDetails": {
            "singleSignOnType": "None",
            "connectionEncryption": connection_encryption,
            "skipTestConnection": skip_test_connection,
            "credentials": {
                "credentialType": "Basic",
                "username": user_name,
                "password": password,
            },
        },
    }

    _base_api(
        request="/v1/connections",
        client="fabric_sp",
        method="post",
        payload=payload,
        status_codes=201,
    )

    print(f"{icons.green_dot} The '{name}' cloud connection has been created.")


def create_on_prem_connection(
    name: str,
    gateway: str | UUID,
    server_name: str,
    database_name: str,
    credentials: str,
    privacy_level: str,
    connection_encryption: str = "NotEncrypted",
    skip_test_connection: bool = False,
):
    """
    Creates an on-premises connection.

    This is a wrapper function for the following API: `Connections - Create Connection <https://learn.microsoft.com/rest/api/fabric/core/connections/create-connection>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name : str
        The name of the connection.
    gateway : str | uuid.UUID
        The name or Id of the gateway.
    server_name : str
        The name of the server.
    database_name : str
        The name of the database.
    user_name : str
        The username.
    password : str
        The password.
    privacy_level : str
        The `privacy level <https://learn.microsoft.com/rest/api/fabric/core/connections/create-connection?tabs=HTTP#privacylevel>`_ of the connection.
    connection_encryption : str, default="NotEncrypted"
        The connection encrpytion.
    skip_test_connection: bool, default=False
        If True, skips the test connection.
    """

    gateway_id = _resolve_gateway_id(gateway)

    payload = {
        "connectivityType": "OnPremisesGateway",
        "gatewayId": gateway_id,
        "displayName": name,
        "connectionDetails": {
            "type": "SQL",
            "creationMethod": "SQL",
            "parameters": [
                {
                    "dataType": "Text",
                    "name": "server",
                    "value": server_name,
                },
                {
                    "dataType": "Text",
                    "name": "database",
                    "value": database_name,
                },
            ],
        },
        "privacyLevel": privacy_level,
        "credentialDetails": {
            "singleSignOnType": "None",
            "connectionEncryption": connection_encryption,
            "skipTestConnection": skip_test_connection,
            "credentials": {
                "credentialType": "Windows",
                "values": [{"gatewayId": gateway_id, "credentials": credentials}],
            },
        },
    }

    _base_api(
        request="/v1/connections",
        client="fabric_sp",
        method="post",
        payload=payload,
        status_codes=201,
    )

    print(f"{icons.green_dot} The '{name}' on-prem connection has been created.")


def create_vnet_connection(
    name: str,
    gateway: str | UUID,
    server_name: str,
    database_name: str,
    user_name: str,
    password: str,
    privacy_level: str,
    connection_encryption: str = "NotEncrypted",
    skip_test_connection: bool = False,
):
    """
    Creates a virtual network gateway connection.

    This is a wrapper function for the following API: `Connections - Create Connection <https://learn.microsoft.com/rest/api/fabric/core/connections/create-connection>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name : str
        The name of the connection.
    gateway : str | uuid.UUID
        The name or Id of the gateway.
    server_name : str
        The name of the server.
    database_name : str
        The name of the database.
    user_name : str
        The username.
    password : str
        The password.
    privacy_level : str
        The `privacy level <https://learn.microsoft.com/rest/api/fabric/core/connections/create-connection?tabs=HTTP#privacylevel>`_ of the connection.
    connection_encryption : str, default="NotEncrypted"
        The connection encrpytion.
    skip_test_connection: bool, default=False
        If True, skips the test connection.
    """

    gateway_id = _resolve_gateway_id(gateway)

    payload = {
        "connectivityType": "VirtualNetworkGateway",
        "gatewayId": gateway_id,
        "displayName": name,
        "connectionDetails": {
            "type": "SQL",
            "creationMethod": "SQL",
            "parameters": [
                {
                    "dataType": "Text",
                    "name": "server",
                    "value": server_name,
                },
                {
                    "dataType": "Text",
                    "name": "database",
                    "value": database_name,
                },
            ],
        },
        "privacyLevel": privacy_level,
        "credentialDetails": {
            "singleSignOnType": "None",
            "connectionEncryption": connection_encryption,
            "skipTestConnection": skip_test_connection,
            "credentials": {
                "credentialType": "Basic",
                "username": user_name,
                "password": password,
            },
        },
    }

    _base_api(
        request="/v1/connections",
        client="fabric_sp",
        method="post",
        payload=payload,
        status_codes=201,
    )

    print(
        f"{icons.green_dot} The '{name}' virtual network gateway connection has been created."
    )



import sempy.fabric as fabric
import pandas as pd
import sempy_labs._icons as icons
from typing import Optional
import base64
import requests
from sempy._utils._log import log
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _decode_b64,
    _base_api,
    resolve_item_id,
    create_item,
)
from sempy.fabric.exceptions import FabricHTTPException
import os
from uuid import UUID

_notebook_prefix = "notebook-content."


def _get_notebook_definition_base(
    notebook_name: str, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_id = resolve_item_id(item=notebook_name, type="Notebook", workspace=workspace)
    result = _base_api(
        request=f"v1/workspaces/{workspace_id}/notebooks/{item_id}/getDefinition",
        method="post",
        lro_return_json=True,
        status_codes=None,
    )

    return pd.json_normalize(result["definition"]["parts"])


def _get_notebook_type(
    notebook_name: str, workspace: Optional[str | UUID] = None
) -> str:

    df_items = _get_notebook_definition_base(
        notebook_name=notebook_name, workspace=workspace
    )

    file_path = df_items[df_items["path"].str.startswith(_notebook_prefix)][
        "path"
    ].iloc[0]

    _, file_extension = os.path.splitext(file_path)

    return file_extension[1:]


def get_notebook_definition(
    notebook_name: str, workspace: Optional[str | UUID] = None, decode: bool = True
) -> str:
    """
    Obtains the notebook definition.

    This is a wrapper function for the following API: `Items - Get Notebook Definition <https://learn.microsoft.com/rest/api/fabric/notebook/items/get-notebook-definition>`_.

    Parameters
    ----------
    notebook_name : str
        The name of the notebook.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    decode : bool, default=True
        If True, decodes the notebook definition file into .ipynb format.
        If False, obtains the notebook definition file in base64 format.

    Returns
    -------
    str
        The notebook definition.
    """

    df_items = _get_notebook_definition_base(
        notebook_name=notebook_name, workspace=workspace
    )
    df_items_filt = df_items[df_items["path"].str.startswith(_notebook_prefix)]
    payload = df_items_filt["payload"].iloc[0]

    if decode:
        result = _decode_b64(payload)
    else:
        result = payload

    return result


@log
def import_notebook_from_web(
    notebook_name: str,
    url: str,
    description: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
    overwrite: bool = False,
):
    """
    Creates a new notebook within a workspace based on a Jupyter notebook hosted in the web.

    Note: When specifying a notebook from GitHub, please use the raw file path. Note that if the non-raw file path is specified, the url will be
    converted to the raw URL as the raw URL is needed to obtain the notebook content.

    Parameters
    ----------
    notebook_name : str
        The name of the notebook to be created.
    url : str
        The url of the Jupyter Notebook (.ipynb)
    description : str, default=None
        The description of the notebook.
        Defaults to None which does not place a description.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    overwrite : bool, default=False
        If set to True, overwrites the existing notebook in the workspace if it exists.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    # Fix links to go to the raw github file
    starting_text = "https://github.com/"
    starting_text_len = len(starting_text)
    if url.startswith(starting_text):
        url = f"https://raw.githubusercontent.com/{url[starting_text_len:]}".replace(
            "/blob/", "/"
        )

    response = requests.get(url)
    if response.status_code != 200:
        raise FabricHTTPException(response)

    dfI = fabric.list_items(workspace=workspace, type="Notebook")
    dfI_filt = dfI[dfI["Display Name"] == notebook_name]
    if len(dfI_filt) == 0:
        create_notebook(
            name=notebook_name,
            notebook_content=response.content,
            workspace=workspace_id,
            description=description,
        )
    elif len(dfI_filt) > 0 and overwrite:
        print(f"{icons.info} Overwrite of notebooks is currently not supported.")
        # update_notebook_definition(
        #    name=notebook_name, notebook_content=response.content, workspace=workspace
        # )
    else:
        raise ValueError(
            f"{icons.red_dot} The '{notebook_name}' already exists within the '{workspace_name}' workspace and 'overwrite' is set to False."
        )


def create_notebook(
    name: str,
    notebook_content: str,
    type: str = "py",
    description: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Creates a new notebook with a definition within a workspace.

    Parameters
    ----------
    name : str
        The name of the notebook to be created.
    notebook_content : str
        The Jupyter notebook content (not in Base64 format).
    type : str, default="py"
        The notebook type.
    description : str, default=None
        The description of the notebook.
        Defaults to None which does not place a description.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    notebook_payload = base64.b64encode(notebook_content).decode("utf-8")
    definition_payload = {
        "format": "ipynb",
        "parts": [
            {
                "path": f"{_notebook_prefix}.{type}",
                "payload": notebook_payload,
                "payloadType": "InlineBase64",
            }
        ],
    }

    create_item(
        name=name,
        type="Notebook",
        workspace=workspace,
        description=description,
        definition=definition_payload,
    )


def update_notebook_definition(
    name: str, notebook_content: str, workspace: Optional[str | UUID] = None
):
    """
    Updates an existing notebook with a new definition.

    Parameters
    ----------
    name : str
        The name of the notebook to be updated.
    notebook_content : str
        The Jupyter notebook content (not in Base64 format).
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    notebook_payload = base64.b64encode(notebook_content)
    item_id = resolve_item_id(item=name, type="Notebook", workspace=workspace)
    type = _get_notebook_type(notebook_name=name, workspace=workspace)

    payload = {
        "definition": {
            "parts": [
                {
                    "path": f"{_notebook_prefix}.{type}",
                    "payload": notebook_payload,
                    "payloadType": "InlineBase64",
                }
            ],
        },
    }

    _base_api(
        request=f"v1/workspaces/{workspace_id}/notebooks/{item_id}/updateDefinition",
        payload=payload,
        method="post",
        lro_return_status_code=True,
        status_codes=None,
    )

    print(
        f"{icons.green_dot} The '{name}' notebook was updated within the '{workspace_name}' workspace."
    )



from sempy_labs.directlake._generate_shared_expression import generate_shared_expression
from sempy_labs._helper_functions import (
    resolve_dataset_name_and_id,
    resolve_workspace_name_and_id,
    resolve_item_name_and_id,
    resolve_lakehouse_name_and_id,
)
from sempy._utils._log import log
from sempy_labs.tom import connect_semantic_model
from typing import Optional
import sempy_labs._icons as icons
from uuid import UUID
import re


def _extract_expression_list(expression):
    """
    Finds the pattern for DL/SQL & DL/OL expressions in the semantic model.
    """

    pattern_sql = r'Sql\.Database\s*\(\s*"([^"]+)"\s*,\s*"([^"]+)"\s*\)'
    pattern_no_sql = r'AzureDataLakeStorage\s*\{\s*"server".*?:\s*onelake\.dfs\.fabric\.microsoft\.com"\s*,\s*"path"\s*:\s*"/([\da-fA-F-]+)\s*/\s*([\da-fA-F-]+)\s*/"\s*\}'

    match_sql = re.search(pattern_sql, expression)
    match_no_sql = re.search(pattern_no_sql, expression)

    result = []
    if match_sql:
        value_1, value_2 = match_sql.groups()
        result = [value_1, value_2, True]
    elif match_no_sql:
        value_1, value_2 = match_no_sql.groups()
        result = [value_1, value_2, False]

    return result


def _get_direct_lake_expressions(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> dict:
    """
    Extracts a dictionary of all Direct Lake expressions from a semantic model.
    """

    from sempy_labs.tom import connect_semantic_model

    result = {}

    with connect_semantic_model(dataset=dataset, workspace=workspace) as tom:
        for e in tom.model.Expressions:
            expr_name = e.Name
            expr = e.Expression

            list_values = _extract_expression_list(expr)
            if list_values:
                result[expr_name] = list_values

    return result


@log
def update_direct_lake_model_lakehouse_connection(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    lakehouse: Optional[str] = None,
    lakehouse_workspace: Optional[str | UUID] = None,
):
    """
    Remaps a Direct Lake semantic model's SQL Endpoint connection to a new lakehouse.

    Parameters
    ----------
    dataset : str | UUID
        Name or ID of the semantic model.
    workspace : str | UUID, default=None
        The Fabric workspace name or ID in which the semantic model exists.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    lakehouse : str, default=None
        The Fabric lakehouse used by the Direct Lake semantic model.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    lakehouse_workspace : str | UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    update_direct_lake_model_connection(
        dataset=dataset,
        workspace=workspace,
        source=lakehouse,
        source_type="Lakehouse",
        source_workspace=lakehouse_workspace,
    )


@log
def update_direct_lake_model_connection(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    source: Optional[str] = None,
    source_type: str = "Lakehouse",
    source_workspace: Optional[str | UUID] = None,
    use_sql_endpoint: bool = True,
):
    """
    Remaps a Direct Lake semantic model's SQL Endpoint connection to a new lakehouse/warehouse.

    Parameters
    ----------
    dataset : str | UUID
        Name or ID of the semantic model.
    workspace : str | UUID, default=None
        The Fabric workspace name or ID in which the semantic model exists.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    source : str, default=None
        The name of the Fabric lakehouse/warehouse used by the Direct Lake semantic model.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    source_type : str, default="Lakehouse"
        The type of source for the Direct Lake semantic model. Valid options: "Lakehouse", "Warehouse".
    source_workspace : str | UUID, default=None
        The Fabric workspace name or ID used by the lakehouse/warehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    use_sql_endpoint : bool, default=True
        If True, the SQL Endpoint will be used for the connection.
        If False, Direct Lake over OneLake will be used.
    """
    if use_sql_endpoint:
        icons.sll_tags.append("UpdateDLConnection_SQL")
    else:
        icons.sll_tags.append("UpdateDLConnection_DLOL")

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    source_type = source_type.capitalize()

    if source_type not in ["Lakehouse", "Warehouse"]:
        raise ValueError(
            f"{icons.red_dot} The 'source_type' must be either 'Lakehouse' or 'Warehouse'."
        )

    if source_workspace is None:
        source_workspace = workspace_name

    if source_type == "Lakehouse":
        (source_name, source_id) = resolve_lakehouse_name_and_id(
            lakehouse=source, workspace=source_workspace
        )
    else:
        (source_name, source_id) = resolve_item_name_and_id(
            item=source, type=source_type, workspace=source_workspace
        )

    shared_expression = generate_shared_expression(
        item_name=source_name,
        item_type=source_type,
        workspace=source_workspace,
        use_sql_endpoint=use_sql_endpoint,
    )

    expression_dict = _get_direct_lake_expressions(dataset=dataset, workspace=workspace)
    expressions = list(expression_dict.keys())

    with connect_semantic_model(
        dataset=dataset_id, readonly=False, workspace=workspace_id
    ) as tom:

        if not tom.is_direct_lake():
            raise ValueError(
                f"{icons.red_dot} The '{dataset_name}' semantic model within the '{workspace_name}' workspace is not in Direct Lake. This function is only applicable to Direct Lake semantic models."
            )

        # Update the single connection expression
        if len(expressions) == 1:
            expr = expressions[0]
            tom.model.Expressions[expr].Expression = shared_expression

            print(
                f"{icons.green_dot} The expression in the '{dataset_name}' semantic model within the '{workspace_name}' workspace has been updated to point to the '{source}' {source_type.lower()} in the '{source_workspace}' workspace."
            )
        else:
            print(
                f"{icons.info} Multiple expressions found in the model. Please use the update_direct_lake_partition_entity function to update specific tables."
            )



import sempy.fabric as fabric
import pandas as pd
import json
import os
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _conv_b64,
    resolve_dataset_name_and_id,
    resolve_item_name_and_id,
    _update_dataframe_datatypes,
    _base_api,
    resolve_item_id,
    get_item_definition,
)
import sempy_labs._icons as icons
from sempy._utils._log import log
from uuid import UUID


def create_report_from_reportjson(
    report: str,
    dataset: str | UUID,
    report_json: dict,
    theme_json: Optional[dict] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Creates a report based on a report.json file (and an optional themes.json file).

    This is a wrapper function for the following API: `Items - Create Report <https://learn.microsoft.com/rest/api/fabric/report/items/create-report>`_.

    Parameters
    ----------
    report : str
        Name of the report.
    dataset : str | uuid.UUID
        Name or ID of the semantic model to connect to the report.
    report_json : dict
        The report.json file to be used to create the report.
    theme_json : dict, default=None
        The theme.json file to be used for the theme of the report.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    dfI = fabric.list_items(workspace=workspace, type="Report")
    dfI_rpt = dfI[dfI["Display Name"] == report]

    if not dfI_rpt.empty:
        print(
            f"{icons.yellow_dot} '{report}' report already exists in the '{workspace_name}' workspace."
        )
        return

    defPBIR = {
        "version": "1.0",
        "datasetReference": {
            "byPath": None,
            "byConnection": {
                "connectionString": None,
                "pbiServiceModelId": None,
                "pbiModelVirtualServerName": "sobe_wowvirtualserver",
                "pbiModelDatabaseName": dataset_id,
                "name": "EntityDataSource",
                "connectionType": "pbiServiceXmlaStyleLive",
            },
        },
    }

    definitionPBIR = _conv_b64(defPBIR)
    payloadReportJson = _conv_b64(report_json)

    request_body = {
        "displayName": report,
        "definition": {
            "parts": [
                {
                    "path": "report.json",
                    "payload": payloadReportJson,
                    "payloadType": "InlineBase64",
                },
                {
                    "path": "definition.pbir",
                    "payload": definitionPBIR,
                    "payloadType": "InlineBase64",
                },
            ]
        },
    }

    if theme_json is not None:
        theme_payload = _conv_b64(theme_json)
        theme_id = theme_json["payload"]["blob"]["displayName"]
        theme_path = f"StaticResources/SharedResources/BaseThemes/{theme_id}.json"

        part = {
            "path": theme_path,
            "payload": theme_payload,
            "payloadType": "InlineBase64",
        }
        request_body["definition"]["parts"].append(part)

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/reports",
        method="post",
        payload=request_body,
        lro_return_status_code=True,
        status_codes=[201, 202],
    )

    print(
        f"{icons.green_dot} Succesfully created the '{report}' report within the '{workspace_name}' workspace."
    )


def update_report_from_reportjson(
    report: str | UUID, report_json: dict, workspace: Optional[str | UUID] = None
):
    """
    Updates a report based on a report.json file.

    This is a wrapper function for the following API: `Items - Update Report Definition <https://learn.microsoft.com/rest/api/fabric/report/items/update-report-definition>`_.

    Parameters
    ----------
    report : str | uuid.UUID
        Name or ID of the report.
    report_json : dict
        The report.json file to be used to update the report.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID in which the report resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    report_id = resolve_item_id(item=report, type="Report", workspace=workspace)

    # Get the existing PBIR file
    df_items = get_report_definition(report=report, workspace=workspace_id)
    df_items_filt = df_items[df_items["path"] == "definition.pbir"]
    rptDefFile = df_items_filt["payload"].iloc[0]
    payloadReportJson = _conv_b64(report_json)

    payload = {
        "definition": {
            "parts": [
                {
                    "path": "report.json",
                    "payload": payloadReportJson,
                    "payloadType": "InlineBase64",
                },
                {
                    "path": "definition.pbir",
                    "payload": rptDefFile,
                    "payloadType": "InlineBase64",
                },
            ]
        }
    }

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/reports/{report_id}/updateDefinition",
        method="post",
        payload=payload,
        lro_return_status_code=True,
        status_codes=None,
    )

    print(
        f"{icons.green_dot} The '{report}' report within the '{workspace_name}' workspace has been successfully updated."
    )


def get_report_definition(
    report: str | UUID,
    workspace: Optional[str | UUID] = None,
    return_dataframe: bool = True,
) -> pd.DataFrame | dict:
    """
    Gets the collection of definition files of a report.

    This is a wrapper function for the following API: `Items - Get Report Definition <https://learn.microsoft.com/rest/api/fabric/report/items/get-report-definition>`_.

    Parameters
    ----------
    report : str | uuid.UUID
        Name or ID of the report.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID in which the report resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    return_dataframe : bool, default=True
        If True, returns a dataframe. If False, returns a json dictionary.

    Returns
    -------
    pandas.DataFrame
        The collection of report definition files within a pandas dataframe.
    """

    return get_item_definition(
        item=report, type="Report", workspace=workspace, return_dataframe=True
    )


@log
def create_model_bpa_report(
    report: Optional[str] = icons.model_bpa_name,
    dataset: Optional[str] = icons.model_bpa_name,
    dataset_workspace: Optional[str | UUID] = None,
):
    """
    Dynamically generates a Best Practice Analyzer report for analyzing semantic models.

    Parameters
    ----------
    report : str, default='ModelBPA'
        Name of the report.
        Defaults to 'ModelBPA'.
    dataset : str, default='ModelBPA'
        Name of the semantic model which feeds this report.
        Defaults to 'ModelBPA'
    dataset_workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID in which the semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    """
    (dataset_workspace_name, dataset_workspace_id) = resolve_workspace_name_and_id(
        dataset_workspace
    )

    dfI = fabric.list_items(workspace=dataset_workspace_id, type="SemanticModel")
    dfI_filt = dfI[dfI["Display Name"] == dataset]

    if len(dfI_filt) == 0:
        raise ValueError(
            f"The '{dataset}' semantic model does not exist within the '{dataset_workspace_name}' workspace."
        )

    dfR = fabric.list_reports(workspace=dataset_workspace_id)
    dfR_filt = dfR[dfR["Name"] == report]
    # dataset_id = resolve_dataset_id(dataset=dataset, workspace=dataset_workspace)

    current_dir = os.path.dirname(__file__)
    # directory_path = os.path.join(current_dir, "_bpareporttemplate")
    # len_dir_path = len(directory_path) + 1

    # request_body = {"displayName": report, "definition": {"parts": []}}

    # def get_all_file_paths(directory):
    #    file_paths = []

    #    for root, directories, files in os.walk(directory):
    #        for filename in files:
    #            full_path = os.path.join(root, filename)
    #            file_paths.append(full_path)

    #    return file_paths

    # all_files = get_all_file_paths(directory_path)

    # for file_path in all_files:
    #    fp = file_path[len_dir_path:]
    #    with open(file_path, "r") as file:
    #        json_file = json.load(file)
    #        if fp == 'definition.pbir':
    #            conn_string = f"Data Source=powerbi://api.powerbi.com/v1.0/myorg/{dataset_workspace};Initial Catalog={dataset};Integrated Security=ClaimsToken"
    #            json_file['datasetReference']['byConnection']['connectionString'] = conn_string
    #            json_file['datasetReference']['byConnection']['pbiModelDatabaseName'] = dataset_id
    #        part = {
    #            "path": fp,
    #            "payload": _conv_b64(json_file),
    #            "payloadType": "InlineBase64",
    #        }

    #    request_body["definition"]["parts"].append(part)

    # _create_report(
    #    report=report,
    #    request_body=request_body,
    #    dataset=dataset,
    #    report_workspace=dataset_workspace,
    #    dataset_workspace=dataset_workspace,
    # )

    json_file_path = os.path.join(current_dir, "_BPAReportTemplate.json")
    with open(json_file_path, "r") as file:
        report_json = json.load(file)

    if len(dfR_filt) > 0:
        update_report_from_reportjson(
            report=report, report_json=report_json, workspace=dataset_workspace_id
        )
    else:
        create_report_from_reportjson(
            report=report,
            dataset=dataset,
            report_json=report_json,
            workspace=dataset_workspace_id,
        )


def _create_report(
    report: str,
    request_body: dict,
    dataset: str,
    dataset_workspace: Optional[str] = None,
    report_workspace: Optional[str] = None,
    overwrite: bool = False,
):

    from sempy_labs.report import report_rebind

    (report_workspace_name, report_workspace_id) = resolve_workspace_name_and_id(
        workspace=report_workspace
    )

    dfR = fabric.list_reports(workspace=report_workspace)
    dfR_filt = dfR[dfR["Name"] == report]

    updated_report = False
    # Create report if it does not exist
    if dfR_filt.empty:
        _base_api(
            request=f"/v1/workspaces/{report_workspace_id}/reports",
            method="post",
            payload=request_body,
            lro_return_status_code=True,
            status_codes=[201, 202],
        )

        print(
            f"{icons.green_dot} The '{report}' report has been created within the '{report_workspace_name}'"
        )
        updated_report = True
    # Update the report if it exists
    elif not dfR_filt.empty and overwrite:
        report_id = dfR_filt["Id"].iloc[0]
        _base_api(
            request=f"/v1/workspaces/{report_workspace_id}/reports/{report_id}/updateDefinition",
            method="post",
            payload=request_body,
            lro_return_status_code=True,
            status_codes=None,
        )
        print(
            f"{icons.green_dot} The '{report}' report has been updated within the '{report_workspace_name}'"
        )
        updated_report = True
    else:
        raise ValueError(
            f"{icons.red_dot} The '{report}' report within the '{report_workspace_name}' workspace already exists and the 'overwrite' parameter was set to False."
        )

    # Rebind the report to the semantic model to make sure it is pointed at the correct semantic model
    if updated_report:
        report_rebind(
            report=report,
            dataset=dataset,
            report_workspace=report_workspace,
            dataset_workspace=dataset_workspace,
        )


def _get_report(
    report: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (report_name, report_id) = resolve_item_name_and_id(
        item=report, type="Report", workspace=workspace
    )

    response = _base_api(
        request=f"v1.0/myorg/groups/{workspace_id}/reports/{report_id}"
    )
    result = response.json()

    new_data = {
        "Id": result.get("id"),
        "Report Type": result.get("reportType"),
        "Name": result.get("name"),
        "Web Url": result.get("webUrl"),
        "Embed Url": result.get("embedUrl"),
        "Is From Pbix": result.get("isFromPbix"),
        "Is Owned By Me": result.get("isOwnedByMe"),
        "Dataset Id": result.get("datasetId"),
        "Dataset Workspace Id": result.get("datasetWorkspaceId"),
        "Users": result.get("users") if result.get("users") is not None else [],
        "Subscriptions": (
            result.get("subscriptions")
            if result.get("subscriptions") is not None
            else []
        ),
    }

    df = pd.DataFrame([new_data])

    column_map = {
        "Is From Pbix": "bool",
        "Is Owned By Me": "bool",
    }
    _update_dataframe_datatypes(dataframe=df, column_map=column_map)

    return df



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    _build_url,
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
    _is_valid_uuid,
)
from uuid import UUID
import sempy_labs._icons as icons


def list_apps(
    top: Optional[int] = 1000,
    skip: Optional[int] = None,
) -> pd.DataFrame:
    """
    Shows a list of apps in the organization.

    This is a wrapper function for the following API: `Admin - Apps GetAppsAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/apps-get-apps-as-admin>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    top : int, default=1000
        Returns only the first n results.
    skip : int, default=None
        Skips the first n results.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of apps in the organization.
    """

    columns = {
        "App Name": "string",
        "App Id": "string",
        "Description": "string",
        "Published By": "string",
        "Last Update": "datetime_coerce",
    }

    df = _create_dataframe(columns=columns)

    params = {}
    url = "/v1.0/myorg/admin/apps"

    params["$top"] = top

    if skip is not None:
        params["$skip"] = skip

    url = _build_url(url, params)
    response = _base_api(request=url, client="fabric_sp")

    rows = []
    for v in response.json().get("value", []):
        rows.append(
            {
                "App Name": v.get("name"),
                "App Id": v.get("id"),
                "Description": v.get("description"),
                "Published By": v.get("publishedBy"),
                "Last Update": v.get("lastUpdate"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


def _resolve_app_id(app: str | UUID) -> str:
    if _is_valid_uuid(app):
        return app
    else:
        df = list_apps()
        df_filt = df[df["App Name"] == app]
        if df_filt.empty:
            raise ValueError(f"{icons.red_dot} The '{app}' app does not exist.")
        return df_filt["App Id"].iloc[0]


def list_app_users(app: str | UUID) -> pd.DataFrame:
    """
    Shows a list of users that have access to the specified app.

    This is a wrapper function for the following API: `Admin - Apps GetAppUsersAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/apps-get-app-users-as-admin>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    app : str | uuid.UUID
        The name or ID of the app.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of users that have access to the specified app.
    """

    app_id = _resolve_app_id(app)

    columns = {
        "User Name": "string",
        "Email Address": "string",
        "App User Access Right": "string",
        "Identifier": "string",
        "Graph Id": "string",
        "Principal Type": "string",
    }

    df = _create_dataframe(columns=columns)

    url = f"/v1.0/myorg/admin/apps/{app_id}/users"
    response = _base_api(request=url, client="fabric_sp")

    rows = []
    for v in response.json().get("value", []):
        rows.append(
            {
                "User Name": v.get("displayName"),
                "Email Address": v.get("emailAddress"),
                "App User Access Right": v.get("appUserAccessRight"),
                "Identifier": v.get("identifier"),
                "Graph Id": v.get("graphId"),
                "Principal Type": v.get("principalType"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df



import sempy.fabric as fabric
from sempy_labs._helper_functions import (
    is_default_semantic_model,
    _get_adls_client,
    resolve_workspace_name_and_id,
    resolve_dataset_name_and_id,
    _update_dataframe_datatypes,
    _base_api,
)
from typing import Optional
import sempy_labs._icons as icons
from sempy._utils._log import log
import pandas as pd
from uuid import UUID


@log
def clear_cache(dataset: str | UUID, workspace: Optional[str | UUID] = None):
    """
    Clears the cache of a semantic model.
    See `here <https://learn.microsoft.com/analysis-services/instances/clear-the-analysis-services-caches?view=asallproducts-allversions>`_ for documentation.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    if is_default_semantic_model(dataset=dataset, workspace=workspace):
        raise ValueError(
            f"{icons.red_dot} Cannot run XMLA operations against a default semantic model. Please choose a different semantic model. "
            "See here for more information: https://learn.microsoft.com/fabric/data-warehouse/semantic-models"
        )

    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    xmla = f"""
            <ClearCache xmlns="http://schemas.microsoft.com/analysisservices/2003/engine">
                <Object>
                    <DatabaseID>{dataset_id}</DatabaseID>
                </Object>
            </ClearCache>
            """
    fabric.execute_xmla(dataset=dataset_id, xmla_command=xmla, workspace=workspace_id)
    print(
        f"{icons.green_dot} Cache cleared for the '{dataset_name}' semantic model within the '{workspace_name}' workspace."
    )


@log
def backup_semantic_model(
    dataset: str | UUID,
    file_path: str,
    allow_overwrite: bool = True,
    apply_compression: bool = True,
    workspace: Optional[str | UUID] = None,
):
    """
    `Backs up <https://learn.microsoft.com/azure/analysis-services/analysis-services-backup>`_ a semantic model to the ADLS Gen2 storage account connected to the workspace.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    file_path : str
        The ADLS Gen2 storage account location in which to backup the semantic model. Always saves within the 'power-bi-backup/<workspace name>' folder.
        Must end in '.abf'.
        Example 1: file_path = 'MyModel.abf'
        Example 2: file_path = 'MyFolder/MyModel.abf'
    allow_overwrite : bool, default=True
        If True, overwrites backup files of the same name. If False, the file you are saving cannot have the same name as a file that already exists in the same location.
    apply_compression : bool, default=True
        If True, compresses the backup file. Compressed backup files save disk space, but require slightly higher CPU utilization.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    if not file_path.endswith(".abf"):
        raise ValueError(
            f"{icons.red_dot} The backup file for restoring must be in the .abf format."
        )

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    tmsl = {
        "backup": {
            "database": dataset_name,
            "file": file_path,
            "allowOverwrite": allow_overwrite,
            "applyCompression": apply_compression,
        }
    }

    fabric.execute_tmsl(script=tmsl, workspace=workspace_id)
    print(
        f"{icons.green_dot} The '{dataset_name}' semantic model within the '{workspace_name}' workspace has been backed up to the '{file_path}' location."
    )


@log
def restore_semantic_model(
    dataset: str,
    file_path: str,
    allow_overwrite: bool = True,
    ignore_incompatibilities: bool = True,
    force_restore: bool = False,
    workspace: Optional[str | UUID] = None,
):
    """
    `Restores <https://learn.microsoft.com/power-bi/enterprise/service-premium-backup-restore-dataset>`_ a semantic model based on a backup (.abf) file
    within the ADLS Gen2 storage account connected to the workspace.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    file_path : str
        The location in which to backup the semantic model. Must end in '.abf'.
        Example 1: file_path = 'MyModel.abf'
        Example 2: file_path = 'MyFolder/MyModel.abf'
    allow_overwrite : bool, default=True
        If True, overwrites backup files of the same name. If False, the file you are saving cannot have the same name as a file that already exists in the same location.
    ignore_incompatibilities : bool, default=True
        If True, ignores incompatibilities between Azure Analysis Services and Power BI Premium.
    force_restore: bool, default=False
        If True, restores the semantic model with the existing semantic model unloaded and offline.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    if not file_path.endswith(".abf"):
        raise ValueError(
            f"{icons.red_dot} The backup file for restoring must be in the .abf format."
        )

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    tmsl = {
        "restore": {
            "database": dataset,
            "file": file_path,
            "allowOverwrite": allow_overwrite,
            "security": "copyAll",
            "ignoreIncompatibilities": ignore_incompatibilities,
        }
    }

    if force_restore:
        tmsl["restore"]["forceRestore"] = force_restore

    fabric.execute_tmsl(script=tmsl, workspace=workspace_id)

    print(
        f"{icons.green_dot} The '{dataset}' semantic model has been restored to the '{workspace_name}' workspace based on the '{file_path}' backup file."
    )


@log
def copy_semantic_model_backup_file(
    source_workspace: str,
    target_workspace: str,
    source_file_name: str,
    target_file_name: str,
    storage_account: str,
    source_file_system: str = "power-bi-backup",
    target_file_system: str = "power-bi-backup",
):
    """
    Copies a semantic model backup file (.abf) from an Azure storage account to another location within the Azure storage account.

    Requirements:
        1. Must have an Azure storage account and connect it to both the source and target workspace.
        2. Must be a 'Storage Blob Data Contributor' for the storage account.
            Steps:
                1. Navigate to the storage account within the Azure Portal
                2. Navigate to 'Access Control (IAM)'
                3. Click '+ Add' -> Add Role Assignment
                4. Search for 'Storage Blob Data Contributor', select it and click 'Next'
                5. Add yourself as a member, click 'Next'
                6. Click 'Review + assign'

    Parameters
    ----------
    source_workspace : str
        The workspace name of the source semantic model backup file.
    target_workspace : str
        The workspace name of the target semantic model backup file destination.
    source_file_name : str
        The name of the source backup file (i.e. MyModel.abf).
    target_file_name : str
        The name of the target backup file (i.e. MyModel.abf).
    storage_account : str
        The name of the storage account.
    source_file_system : str, default="power-bi-backup"
        The container in which the source backup file is located.
    target_file_system : str, default="power-bi-backup"
        The container in which the target backup file will be saved.
    """

    suffix = ".abf"

    if not source_file_name.endswith(suffix):
        source_file_name = f"{source_file_name}{suffix}"
    if not target_file_name.endswith(suffix):
        target_file_name = f"{target_file_name}{suffix}"

    source_path = f"/{source_workspace}/{source_file_name}"
    target_path = f"/{target_workspace}/{target_file_name}"

    client = _get_adls_client(account_name=storage_account)

    source_file_system_client = client.get_file_system_client(
        file_system=source_file_system
    )
    destination_file_system_client = client.get_file_system_client(
        file_system=target_file_system
    )

    source_file_client = source_file_system_client.get_file_client(source_path)
    destination_file_client = destination_file_system_client.get_file_client(
        target_path
    )

    download = source_file_client.download_file()
    file_content = download.readall()

    # Upload the content to the destination file
    destination_file_client.create_file()  # Create the destination file
    destination_file_client.append_data(
        data=file_content, offset=0, length=len(file_content)
    )
    destination_file_client.flush_data(len(file_content))

    print(
        f"{icons.green_dot} The backup file of the '{source_file_name}' semantic model from the '{source_workspace}' workspace has been copied as the '{target_file_name}' semantic model backup file within the '{target_workspace}'."
    )


@log
def list_backups(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows a list of backup files contained within a workspace's ADLS Gen2 storage account.
    Requirement: An ADLS Gen2 storage account must be `connected to the workspace <https://learn.microsoft.com/power-bi/transform-model/dataflows/dataflows-azure-data-lake-storage-integration#connect-to-an-azure-data-lake-gen-2-at-a-workspace-level>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of backup files contained within a workspace's ADLS Gen2 storage account.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    response = _base_api(
        request=f"/v1.0/myorg/resources?resourceType=StorageAccount&folderObjectId={workspace_id}"
    )

    v = response.json().get("value", [])
    if not v:
        raise ValueError(
            f"{icons.red_dot} A storage account is not associated with the '{workspace_name}' workspace."
        )
    storage_account = v[0]["resourceName"]

    df = list_storage_account_files(storage_account=storage_account)
    colName = "Storage Account Name"
    df.insert(0, colName, df.pop(colName))

    return df


@log
def list_storage_account_files(
    storage_account: str, container: str = "power-bi-backup"
) -> pd.DataFrame:
    """
    Shows a list of files within an ADLS Gen2 storage account.

    Parameters
    ----------
    storage_account: str
        The name of the ADLS Gen2 storage account.
    container : str, default='power-bi-backup'
        The name of the container.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of files contained within an ADLS Gen2 storage account.
    """

    df = pd.DataFrame(
        columns=[
            "File Path",
            "File Size",
            "Creation Time",
            "Last Modified",
            "Expiry Time",
            "Encryption Scope",
        ]
    )

    client = _get_adls_client(storage_account)
    fs = client.get_file_system_client(container)

    for x in list(fs.get_paths()):
        if not x.is_directory:
            new_data = {
                "File Path": x.name,
                "File Size": x.content_length,
                "Creation Time": x.creation_time,
                "Last Modified": x.last_modified,
                "Expiry Time": x.expiry_time,
                "Encryption Scope": x.encryption_scope,
            }

            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    column_map = {
        "File Size": "int",
    }

    _update_dataframe_datatypes(dataframe=df, column_map=column_map)

    return df



from typing import Optional
from uuid import UUID
from sempy_labs._helper_functions import (
    resolve_dataset_id,
    _get_fabric_context_setting,
    resolve_workspace_id,
)
from sempy._utils._log import log
import gzip
import base64
import urllib.parse


@log
def generate_dax_query_view_url(
    dataset: str | UUID, dax_string: str, workspace: Optional[str | UUID] = None
):
    """
    Prints a URL based on query provided. This URL opens `DAX query view <https://learn.microsoft.com/power-bi/transform-model/dax-query-view>`_ in the Power BI service, connected to the semantic model and using the query provided.

    Parameters
    ----------
    dataset : str | uuid.UUID
        The semantic model name or ID.
    dax_string : str
        The DAX query string.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    workspace_id = resolve_workspace_id(workspace=workspace)
    dataset_id = resolve_dataset_id(dataset=dataset, workspace=workspace_id)

    prefix = _get_fabric_context_setting(name="spark.trident.pbienv").lower()

    if prefix == "prod":
        prefix = "app"

    def gzip_base64_urlsafe(input_string):
        # Compress the string with gzip
        compressed_data = gzip.compress(input_string.encode("utf-8"))

        # Encode the compressed data in base64
        base64_data = base64.b64encode(compressed_data)

        # Make the base64 string URL-safe
        urlsafe_data = urllib.parse.quote_plus(base64_data.decode("utf-8"))

        return urlsafe_data

    formatted_query = gzip_base64_urlsafe(dax_string)

    url = f"https://{prefix}.powerbi.com/groups/{workspace_id}/modeling/{dataset_id}/daxQueryView?query={formatted_query}"

    print(url)



import sempy.fabric as fabric
import re
import json
import base64
import time
import uuid
from sempy.fabric.exceptions import FabricHTTPException, WorkspaceNotFoundException
import pandas as pd
from functools import wraps
import datetime
from typing import Optional, Tuple, List
from uuid import UUID
import sempy_labs._icons as icons
from azure.core.credentials import TokenCredential, AccessToken
import urllib.parse
import numpy as np
from IPython.display import display, HTML
import requests
import sempy_labs._authentication as auth


def _build_url(url: str, params: dict) -> str:
    """
    Build the url with a list of parameters
    """
    url_parts = list(urllib.parse.urlparse(url))
    url_parts[4] = urllib.parse.urlencode(params)
    url = urllib.parse.urlunparse(url_parts)

    return url


def _encode_user(user: str) -> str:

    return urllib.parse.quote(user, safe="@")


def create_abfss_path(
    lakehouse_id: UUID,
    lakehouse_workspace_id: UUID,
    delta_table_name: Optional[str] = None,
    schema: Optional[str] = None,
) -> str:
    """
    Creates an abfss path for a delta table in a Fabric lakehouse.

    Parameters
    ----------
    lakehouse_id : uuid.UUID
        ID of the Fabric lakehouse.
    lakehouse_workspace_id : uuid.UUID
        ID of the Fabric workspace.
    delta_table_name : str, default=None
        Name of the delta table name.
    schema : str, default=None
        The schema of the delta table.

    Returns
    -------
    str
        An abfss path which can be used to save/reference a delta table in a Fabric lakehouse or lakehouse.
    """

    fp = _get_default_file_path()
    path = f"abfss://{lakehouse_workspace_id}@{fp}/{lakehouse_id}"

    if delta_table_name is not None:
        if schema is not None:
            path += f"/{schema}"
        path += f"/Tables/{delta_table_name}"

    return path


def _get_default_file_path() -> str:

    default_file_storage = _get_fabric_context_setting(name="fs.defaultFS")

    return default_file_storage.split("@")[-1][:-1]


def _split_abfss_path(path: str) -> Tuple[UUID, UUID, str]:

    parsed_url = urllib.parse.urlparse(path)

    workspace_id = parsed_url.netloc.split("@")[0]
    item_id = parsed_url.path.lstrip("/").split("/")[0]
    delta_table_name = parsed_url.path.split("/")[-1]

    return workspace_id, item_id, delta_table_name


def format_dax_object_name(table: str, column: str) -> str:
    """
    Formats a table/column combination to the 'Table Name'[Column Name] format.

    Parameters
    ----------
    table : str
        The name of the table.
    column : str
        The name of the column.

    Returns
    -------
    str
        The fully qualified object name.
    """

    return "'" + table + "'[" + column + "]"


def create_relationship_name(
    from_table: str, from_column: str, to_table: str, to_column: str
) -> str:
    """
    Formats a relationship's table/columns into a fully qualified name.

    Parameters
    ----------
    from_table : str
        The name of the table on the 'from' side of the relationship.
    from_column : str
        The name of the column on the 'from' side of the relationship.
    to_table : str
        The name of the table on the 'to' side of the relationship.
    to_column : str
        The name of the column on the 'to' side of the relationship.

    Returns
    -------
    str
        The fully qualified relationship name.
    """

    return (
        format_dax_object_name(from_table, from_column)
        + " -> "
        + format_dax_object_name(to_table, to_column)
    )


def resolve_report_id(
    report: str | UUID, workspace: Optional[str | UUID] = None
) -> UUID:
    """
    Obtains the ID of the Power BI report.

    Parameters
    ----------
    report : str | uuid.UUID
        The name or ID of the Power BI report.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    uuid.UUID
        The ID of the Power BI report.
    """

    return resolve_item_id(item=report, type="Report", workspace=workspace)


def resolve_report_name(report_id: UUID, workspace: Optional[str | UUID] = None) -> str:
    """
    Obtains the name of the Power BI report.

    Parameters
    ----------
    report_id : uuid.UUID
        The name of the Power BI report.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        The name of the Power BI report.
    """

    return resolve_item_name(item_id=report_id, workspace=workspace)


def delete_item(
    item: str | UUID, type: str, workspace: Optional[str | UUID] = None
) -> None:
    """
    Deletes an item from a Fabric workspace.

    Parameters
    ----------
    item : str | uuid.UUID
        The name or ID of the item to be deleted.
    type : str
        The type of the item to be deleted.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    from sempy_labs._utils import item_types

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (item_name, item_id) = resolve_item_name_and_id(item, type, workspace_id)
    item_type = item_types.get(type)[0].lower()

    fabric.delete_item(item_id=item_id, workspace=workspace_id)

    print(
        f"{icons.green_dot} The '{item_name}' {item_type} has been successfully deleted from the '{workspace_name}' workspace."
    )


def create_item(
    name: str,
    type: str,
    description: Optional[str] = None,
    definition: Optional[dict] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Creates an item in a Fabric workspace.

    Parameters
    ----------
    name : str
        The name of the item to be created.
    type : str
        The type of the item to be created.
    description : str, default=None
        A description of the item to be created.
    definition : dict, default=None
        The definition of the item to be created.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """
    from sempy_labs._utils import item_types

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_type = item_types.get(type)[0].lower()
    item_type_url = item_types.get(type)[1]

    payload = {
        "displayName": name,
    }
    if description:
        payload["description"] = description
    if definition:
        payload["definition"] = definition

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/{item_type_url}",
        method="post",
        payload=payload,
        status_codes=[201, 202],
        lro_return_status_code=True,
    )
    print(
        f"{icons.green_dot} The '{name}' {item_type} has been successfully created within the in the '{workspace_name}' workspace."
    )


def get_item_definition(
    item: str | UUID,
    type: str,
    workspace: Optional[str | UUID] = None,
    format: Optional[str] = None,
    return_dataframe: bool = True,
    decode: bool = True,
):

    from sempy_labs._utils import item_types

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_id = resolve_item_id(item, type, workspace_id)
    item_type_url = item_types.get(type)[1]
    path = item_types.get(type)[2]

    url = f"/v1/workspaces/{workspace_id}/{item_type_url}/{item_id}/getDefinition"
    if format:
        url += f"?format={format}"

    result = _base_api(
        request=url,
        method="post",
        status_codes=None,
        lro_return_json=True,
    )

    if return_dataframe:
        return pd.json_normalize(result["definition"]["parts"])

    value = next(
        p.get("payload") for p in result["definition"]["parts"] if p.get("path") == path
    )
    if decode:
        json.loads(_decode_b64(value))
    else:
        return value


def resolve_item_id(
    item: str | UUID, type: Optional[str] = None, workspace: Optional[str | UUID] = None
) -> UUID:

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_id = None

    if _is_valid_uuid(item):
        # Check (optional)
        item_id = item
        try:
            _base_api(
                request=f"/v1/workspaces/{workspace_id}/items/{item_id}",
                client="fabric_sp",
            )
        except FabricHTTPException:
            raise ValueError(
                f"{icons.red_dot} The '{item_id}' item was not found in the '{workspace_name}' workspace."
            )
    else:
        if type is None:
            raise ValueError(
                f"{icons.red_dot} The 'type' parameter is required if specifying an item name."
            )
        responses = _base_api(
            request=f"/v1/workspaces/{workspace_id}/items?type={type}",
            client="fabric_sp",
            uses_pagination=True,
        )
        for r in responses:
            for v in r.get("value", []):
                display_name = v.get("displayName")
                if display_name == item:
                    item_id = v.get("id")
                    break

    if item_id is None:
        raise ValueError(
            f"{icons.red_dot} There's no item '{item}' of type '{type}' in the '{workspace_name}' workspace."
        )

    return item_id


def resolve_item_name_and_id(
    item: str | UUID, type: Optional[str] = None, workspace: Optional[str | UUID] = None
) -> Tuple[str, UUID]:

    workspace_id = resolve_workspace_id(workspace)
    item_id = resolve_item_id(item=item, type=type, workspace=workspace_id)
    item_name = (
        _base_api(
            request=f"/v1/workspaces/{workspace_id}/items/{item_id}", client="fabric_sp"
        )
        .json()
        .get("displayName")
    )

    return item_name, item_id


def resolve_item_name(item_id: UUID, workspace: Optional[str | UUID] = None) -> str:

    workspace_id = resolve_workspace_id(workspace)
    try:
        item_name = (
            _base_api(
                request=f"/v1/workspaces/{workspace_id}/items/{item_id}",
                client="fabric_sp",
            )
            .json()
            .get("displayName")
        )
    except FabricHTTPException:
        raise ValueError(
            f"{icons.red_dot} The '{item_id}' item was not found in the '{workspace_id}' workspace."
        )

    return item_name


def resolve_lakehouse_name_and_id(
    lakehouse: Optional[str | UUID] = None, workspace: Optional[str | UUID] = None
) -> Tuple[str, UUID]:

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    type = "Lakehouse"

    if lakehouse is None:
        lakehouse_id = _get_fabric_context_setting(name="trident.lakehouse.id")
        if lakehouse_id == "":
            raise ValueError(
                f"{icons.red_dot} Cannot resolve a lakehouse. Please enter a valid lakehouse or make sure a lakehouse is attached to the notebook."
            )
        (lakehouse_name, lakehouse_id) = resolve_item_name_and_id(
            item=lakehouse_id, type=type, workspace=workspace_id
        )

    else:
        (lakehouse_name, lakehouse_id) = resolve_item_name_and_id(
            item=lakehouse, type=type, workspace=workspace_id
        )

    return lakehouse_name, lakehouse_id


def resolve_dataset_name_and_id(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> Tuple[str, UUID]:

    (dataset_name, dataset_id) = resolve_item_name_and_id(
        item=dataset, type="SemanticModel", workspace=workspace
    )

    return dataset_name, dataset_id


def resolve_dataset_id(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> UUID:
    """
    Obtains the ID of the semantic model.

    Parameters
    ----------
    dataset : str | uuid.UUID
        The name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    uuid.UUID
        The ID of the semantic model.
    """

    return resolve_item_id(item=dataset, type="SemanticModel", workspace=workspace)


def resolve_dataset_name(
    dataset_id: UUID, workspace: Optional[str | UUID] = None
) -> str:
    """
    Obtains the name of the semantic model.

    Parameters
    ----------
    dataset_id : uuid.UUID
        The name of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        The name of the semantic model.
    """

    return resolve_item_name(item_id=dataset_id, workspace=workspace)


def resolve_lakehouse_name(
    lakehouse_id: Optional[UUID] = None, workspace: Optional[str | UUID] = None
) -> str:
    """
    Obtains the name of the Fabric lakehouse.

    Parameters
    ----------
    lakehouse_id : uuid.UUID, default=None
        The name of the Fabric lakehouse.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        The name of the Fabric lakehouse.
    """

    if lakehouse_id is None:
        lakehouse_id = _get_fabric_context_setting(name="trident.lakehouse.id")
        if lakehouse_id == "":
            raise ValueError(
                f"{icons.red_dot} Cannot resolve a lakehouse. Please enter a valid lakehouse or make sure a lakehouse is attached to the notebook."
            )

    return resolve_item_name(item_id=lakehouse_id, workspace=workspace)


def resolve_lakehouse_id(
    lakehouse: Optional[str | UUID] = None, workspace: Optional[str | UUID] = None
) -> UUID:
    """
    Obtains the ID of the Fabric lakehouse.

    Parameters
    ----------
    lakehouse : str | uuid.UUID, default=None
        The name or ID of the Fabric lakehouse.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    uuid.UUID
        The ID of the Fabric lakehouse.
    """

    if lakehouse is None:
        lakehouse_id = _get_fabric_context_setting(name="trident.lakehouse.id")
        if lakehouse_id == "":
            raise ValueError(
                f"{icons.red_dot} Cannot resolve a lakehouse. Please enter a valid lakehouse or make sure a lakehouse is attached to the notebook."
            )
    else:
        lakehouse_id = resolve_item_id(
            item=lakehouse, type="Lakehouse", workspace=workspace
        )

    return lakehouse_id


def get_direct_lake_sql_endpoint(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> UUID:
    """
    Obtains the SQL Endpoint ID of the semantic model.

    Parameters
    ----------
    dataset : str | uuid.UUID
        The name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    uuid.UUID
        The ID of SQL Endpoint.
    """

    from sempy_labs.tom import connect_semantic_model

    # dfP = fabric.list_partitions(dataset=dataset, workspace=workspace)
    # dfP_filt = dfP[dfP["Mode"] == "DirectLake"]

    # if len(dfP_filt) == 0:
    #    raise ValueError(
    #        f"The '{dataset}' semantic model in the '{workspace}' workspace is not in Direct Lake mode."
    #    )

    with connect_semantic_model(
        dataset=dataset, readonly=True, workspace=workspace
    ) as tom:
        sqlEndpointId = None
        for e in tom.model.Expressions:
            if e.Name == "DatabaseQuery":
                expr = e.Expression
                matches = re.findall(r'"([^"]+)"', expr)
                sqlEndpointId = matches[1]

        if sqlEndpointId is None:
            raise ValueError(f"{icons.red_dot} SQL Endpoint not found.")

        return sqlEndpointId


def generate_embedded_filter(filter: str) -> str:
    """
    Converts the filter expression to a filter expression which can be used by a Power BI embedded URL.

    Parameters
    ----------
    filter : str
        The filter expression for an embedded Power BI report.

    Returns
    -------
    str
        A filter expression usable by a Power BI embedded URL.
    """

    pattern = r"'[^']+'\[[^\[]+\]"
    matches = re.findall(pattern, filter)
    for match in matches:
        matchReplace = (
            match.replace("'", "")
            .replace("[", "/")
            .replace("]", "")
            .replace(" ", "_x0020_")
            .replace("@", "_00x40_")
            .replace("+", "_0x2B_")
            .replace("{", "_007B_")
            .replace("}", "_007D_")
        )
        filter = filter.replace(match, matchReplace)

    pattern = r"\[[^\[]+\]"
    matches = re.findall(pattern, filter)
    for match in matches:
        matchReplace = (
            match.replace("'", "")
            .replace("[", "/")
            .replace("]", "")
            .replace(" ", "_x0020_")
            .replace("@", "_00x40_")
            .replace("+", "_0x2B_")
            .replace("{", "_007B_")
            .replace("}", "_007D_")
        )
        filter = filter.replace(match, matchReplace)

    revised_filter = (
        filter.replace("<=", "le")
        .replace(">=", "ge")
        .replace("<>", "ne")
        .replace("!=", "ne")
        .replace("==", "eq")
        .replace("=", "eq")
        .replace("<", "lt")
        .replace(">", "gt")
        .replace(" && ", " and ")
        .replace(" & ", " and ")
        .replace(" || ", " or ")
        .replace(" | ", " or ")
        .replace("{", "(")
        .replace("}", ")")
    )

    return revised_filter


def save_as_delta_table(
    dataframe,
    delta_table_name: str,
    write_mode: str,
    merge_schema: bool = False,
    schema: Optional[dict] = None,
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Saves a pandas dataframe as a delta table in a Fabric lakehouse.

    Parameters
    ----------
    dataframe : pandas.DataFrame
        The dataframe to be saved as a delta table.
    delta_table_name : str
        The name of the delta table.
    write_mode : str
        The write mode for the save operation. Options: 'append', 'overwrite'.
    merge_schema : bool, default=False
        Merges the schemas of the dataframe to the delta table.
    schema : dict, default=None
        A dictionary showing the schema of the columns and their data types.
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    from pyspark.sql.types import (
        StringType,
        IntegerType,
        FloatType,
        DateType,
        StructType,
        StructField,
        BooleanType,
        LongType,
        DoubleType,
        TimestampType,
    )

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id(
        lakehouse=lakehouse, workspace=workspace_id
    )

    write_modes = ["append", "overwrite"]
    write_mode = write_mode.lower()

    if write_mode not in write_modes:
        raise ValueError(
            f"{icons.red_dot} Invalid 'write_type' parameter. Choose from one of the following values: {write_modes}."
        )

    if " " in delta_table_name:
        raise ValueError(
            f"{icons.red_dot} Invalid 'delta_table_name'. Delta tables in the lakehouse cannot have spaces in their names."
        )

    spark = _create_spark_session()

    type_mapping = {
        "string": StringType(),
        "str": StringType(),
        "integer": IntegerType(),
        "int": IntegerType(),
        "float": FloatType(),
        "date": DateType(),
        "bool": BooleanType(),
        "boolean": BooleanType(),
        "long": LongType(),
        "double": DoubleType(),
        "timestamp": TimestampType(),
    }

    if isinstance(dataframe, pd.DataFrame):
        dataframe.columns = [col.replace(" ", "_") for col in dataframe.columns]
        if schema is None:
            spark_df = spark.createDataFrame(dataframe)
        else:
            schema_map = StructType(
                [
                    StructField(column_name, type_mapping[data_type], True)
                    for column_name, data_type in schema.items()
                ]
            )
            spark_df = spark.createDataFrame(dataframe, schema_map)
    else:
        for col_name in dataframe.columns:
            new_name = col_name.replace(" ", "_")
            dataframe = dataframe.withColumnRenamed(col_name, new_name)
        spark_df = dataframe

    filePath = create_abfss_path(
        lakehouse_id=lakehouse_id,
        lakehouse_workspace_id=workspace_id,
        delta_table_name=delta_table_name,
    )

    if merge_schema:
        spark_df.write.mode(write_mode).format("delta").option(
            "mergeSchema", "true"
        ).save(filePath)
    else:
        spark_df.write.mode(write_mode).format("delta").save(filePath)
    print(
        f"{icons.green_dot} The dataframe has been saved as the '{delta_table_name}' table in the '{lakehouse_name}' lakehouse within the '{workspace_name}' workspace."
    )


def language_validate(language: str):
    """
    Validateds that the language specified exists within the supported langauges.

    Parameters
    ----------
    language : str
        The language code.

    Returns
    -------
    bool
        A True/False indication as to whether the language code is supported.
    """

    url = "https://learn.microsoft.com/azure/ai-services/translator/language-support"

    tables = pd.read_html(url)
    df = tables[0]

    df_filt = df[df["Language code"] == language]

    df_filt2 = df[df["Language"] == language.capitalize()]

    if len(df_filt) == 1:
        lang = df_filt["Language"].iloc[0]
    elif len(df_filt2) == 1:
        lang = df_filt2["Language"].iloc[0]
    else:
        raise ValueError(
            f"{icons.red_dot} The '{language}' language is not a valid language code. Please refer to this link for a list of valid language codes: {url}."
        )

    return lang


def resolve_workspace_id(
    workspace: Optional[str | UUID] = None,
) -> UUID:
    if workspace is None:
        workspace_id = _get_fabric_context_setting(name="trident.workspace.id")
    elif _is_valid_uuid(workspace):
        # Check (optional)
        workspace_id = workspace
        try:
            _base_api(request=f"/v1/workspaces/{workspace_id}", client="fabric_sp")
        except FabricHTTPException:
            raise ValueError(
                f"{icons.red_dot} The '{workspace_id}' workspace was not found."
            )
    else:
        responses = _base_api(
            request="/v1/workspaces", client="fabric_sp", uses_pagination=True
        )
        workspace_id = None
        for r in responses:
            for v in r.get("value", []):
                display_name = v.get("displayName")
                if display_name == workspace:
                    workspace_id = v.get("id")
                    break

    if workspace_id is None:
        raise WorkspaceNotFoundException(workspace)

    return workspace_id


def resolve_workspace_name(workspace_id: Optional[UUID] = None) -> str:

    if workspace_id is None:
        workspace_id = _get_fabric_context_setting(name="trident.workspace.id")

    try:
        response = _base_api(
            request=f"/v1/workspaces/{workspace_id}", client="fabric_sp"
        ).json()
    except FabricHTTPException:
        raise ValueError(
            f"{icons.red_dot} The '{workspace_id}' workspace was not found."
        )

    return response.get("displayName")


def resolve_workspace_name_and_id(
    workspace: Optional[str | UUID] = None,
) -> Tuple[str, str]:
    """
    Obtains the name and ID of the Fabric workspace.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str, uuid.UUID
        The name and ID of the Fabric workspace.
    """

    if workspace is None:
        workspace_id = _get_fabric_context_setting(name="trident.workspace.id")
        workspace_name = resolve_workspace_name(workspace_id)
    elif _is_valid_uuid(workspace):
        workspace_id = workspace
        workspace_name = resolve_workspace_name(workspace_id)
    else:
        responses = _base_api(
            request="/v1/workspaces", client="fabric_sp", uses_pagination=True
        )
        workspace_id = None
        workspace_name = None
        for r in responses:
            for v in r.get("value", []):
                display_name = v.get("displayName")
                if display_name == workspace:
                    workspace_name = workspace
                    workspace_id = v.get("id")
                    break

    if workspace_name is None or workspace_id is None:
        raise WorkspaceNotFoundException(workspace)

    return workspace_name, workspace_id


def resolve_item_id(
    item: str | UUID, type: Optional[str] = None, workspace: Optional[str | UUID] = None
) -> UUID:

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_id = None

    if _is_valid_uuid(item):
        # Check (optional)
        item_id = item
        try:
            _base_api(
                request=f"/v1/workspaces/{workspace_id}/items/{item_id}",
                client="fabric_sp",
            )
        except FabricHTTPException:
            raise ValueError(
                f"{icons.red_dot} The '{item_id}' item was not found in the '{workspace_name}' workspace."
            )
    else:
        if type is None:
            raise ValueError(
                f"{icons.red_dot} The 'type' parameter is required if specifying an item name."
            )
        responses = _base_api(
            request=f"/v1/workspaces/{workspace_id}/items?type={type}",
            client="fabric_sp",
            uses_pagination=True,
        )
        for r in responses:
            for v in r.get("value", []):
                display_name = v.get("displayName")
                if display_name == item:
                    item_id = v.get("id")
                    break

    if item_id is None:
        raise ValueError(
            f"{icons.red_dot} There's no item '{item}' of type '{type}' in the '{workspace_name}' workspace."
        )

    return item_id


def resolve_item_name_and_id(
    item: str | UUID, type: Optional[str] = None, workspace: Optional[str | UUID] = None
) -> Tuple[str, UUID]:

    workspace_id = resolve_workspace_id(workspace)
    item_id = resolve_item_id(item=item, type=type, workspace=workspace_id)
    item_name = (
        _base_api(
            request=f"/v1/workspaces/{workspace_id}/items/{item_id}", client="fabric_sp"
        )
        .json()
        .get("displayName")
    )

    return item_name, item_id


def resolve_item_name(item_id: UUID, workspace: Optional[str | UUID] = None) -> str:

    workspace_id = resolve_workspace_id(workspace)
    try:
        item_name = (
            _base_api(
                request=f"/v1/workspaces/{workspace_id}/items/{item_id}",
                client="fabric_sp",
            )
            .json()
            .get("displayName")
        )
    except FabricHTTPException:
        raise ValueError(
            f"{icons.red_dot} The '{item_id}' item was not found in the '{workspace_id}' workspace."
        )

    return item_name


def resolve_lakehouse_name_and_id(
    lakehouse: Optional[str | UUID] = None, workspace: Optional[str | UUID] = None
) -> Tuple[str, UUID]:

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    type = "Lakehouse"

    if lakehouse is None:
        lakehouse_id = fabric.get_lakehouse_id()
        lakehouse_name = fabric.resolve_item_name(
            item_id=lakehouse_id, type=type, workspace=workspace_id
        )
    elif _is_valid_uuid(lakehouse):
        lakehouse_id = lakehouse
        lakehouse_name = fabric.resolve_item_name(
            item_id=lakehouse_id, type=type, workspace=workspace_id
        )
    else:
        lakehouse_name = lakehouse
        lakehouse_id = fabric.resolve_item_id(
            item_name=lakehouse, type=type, workspace=workspace_id
        )

    return lakehouse_name, lakehouse_id


def _extract_json(dataframe: pd.DataFrame) -> dict:

    payload = dataframe["payload"].iloc[0]
    json_file = _decode_b64(payload)

    return json.loads(json_file)


def _conv_b64(file):

    loadJson = json.dumps(file)
    f = base64.b64encode(loadJson.encode("utf-8")).decode("utf-8")

    return f


def _decode_b64(file, format: Optional[str] = "utf-8"):

    return base64.b64decode(file).decode(format)


def is_default_semantic_model(
    dataset: str, workspace: Optional[str | UUID] = None
) -> bool:
    """
    Identifies whether a semantic model is a default semantic model.

    Parameters
    ----------
    dataset : str
        The name of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    bool
        A True/False value indicating whether the semantic model is a default semantic model.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    dfI = fabric.list_items(workspace=workspace_id)
    filtered_df = dfI.groupby("Display Name").filter(
        lambda x: set(["Warehouse", "SemanticModel"]).issubset(set(x["Type"]))
        or set(["Lakehouse", "SemanticModel"]).issubset(set(x["Type"]))
    )
    default_semantic_models = filtered_df["Display Name"].unique().tolist()

    return dataset in default_semantic_models


def resolve_item_type(item_id: UUID, workspace: Optional[str | UUID] = None) -> str:
    """
    Obtains the item type for a given Fabric Item Id within a Fabric workspace.

    Parameters
    ----------
    item_id : uuid.UUID
        The item/artifact Id.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        The item type for the item Id.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    dfI = fabric.list_items(workspace=workspace_id)
    dfI_filt = dfI[dfI["Id"] == item_id]

    if dfI_filt.empty:
        raise ValueError(
            f"{icons.red_dot} Invalid 'item_id' parameter. The '{item_id}' item was not found in the '{workspace_name}' workspace."
        )
    return dfI_filt["Type"].iloc[0]


def resolve_dataset_from_report(
    report: str | UUID, workspace: Optional[str | UUID] = None
) -> Tuple[UUID, str, UUID, str]:
    """
    Obtains the basic semantic model properties from which the report's data is sourced.

    Parameters
    ----------
    report : str | uuid.UUID
        The name or ID of the Power BI report.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    Tuple[uuid.UUID, str, uuid.UUID, str]
        The semantic model UUID, semantic model name, semantic model workspace UUID, semantic model workspace name
    """

    from sempy_labs.report._generate_report import _get_report

    dfR = _get_report(report=report, workspace=workspace)
    dataset_id = dfR["Dataset Id"].iloc[0]
    dataset_workspace_id = dfR["Dataset Workspace Id"].iloc[0]
    dataset_workspace = resolve_workspace_name(workspace_id=dataset_workspace_id)
    dataset_name = resolve_dataset_name(
        dataset_id=dataset_id, workspace=dataset_workspace
    )

    return dataset_id, dataset_name, dataset_workspace_id, dataset_workspace


def _add_part(target_dict, path, payload):

    part = {"path": path, "payload": payload, "payloadType": "InlineBase64"}

    target_dict["definition"]["parts"].append(part)


def resolve_workspace_capacity(
    workspace: Optional[str | UUID] = None,
) -> Tuple[UUID, str]:
    """
    Obtains the capacity Id and capacity name for a given workspace.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or UUID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    Tuple[uuid.UUID, str]
        capacity Id; capacity came.
    """
    from sempy_labs._capacities import list_capacities

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    filter_condition = urllib.parse.quote(workspace_id)
    dfW = fabric.list_workspaces(filter=f"id eq '{filter_condition}'")
    capacity_id = dfW["Capacity Id"].iloc[0]
    dfC = list_capacities()
    dfC_filt = dfC[dfC["Id"] == capacity_id]
    if len(dfC_filt) == 1:
        capacity_name = dfC_filt["Display Name"].iloc[0]
    else:
        capacity_name = None

    return capacity_id, capacity_name


def get_capacity_id(workspace: Optional[str | UUID] = None) -> UUID:
    """
    Obtains the Capacity Id for a given workspace.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    uuid.UUID
        The capacity Id.
    """

    if workspace is None:
        capacity_id = _get_fabric_context_setting(name="trident.capacity.id")
    else:
        (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
        filter_condition = urllib.parse.quote(workspace_id)
        dfW = fabric.list_workspaces(filter=f"id eq '{filter_condition}'")
        if len(dfW) == 0:
            raise ValueError(f"{icons.red_dot} The '{workspace_name}' does not exist'.")

        capacity_id = dfW["Capacity Id"].iloc[0]

    return capacity_id


def get_capacity_name(workspace: Optional[str | UUID] = None) -> str:
    """
    Obtains the capacity name for a given workspace.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        The capacity name.
    """

    from sempy_labs._capacities import list_capacities

    capacity_id = get_capacity_id(workspace)
    dfC = list_capacities()
    dfC_filt = dfC[dfC["Id"] == capacity_id]
    if dfC_filt.empty:
        raise ValueError(
            f"{icons.red_dot} The '{capacity_id}' capacity Id does not exist."
        )

    return dfC_filt["Display Name"].iloc[0]


def resolve_capacity_name(capacity_id: Optional[UUID] = None) -> str:
    """
    Obtains the capacity name for a given capacity Id.

    Parameters
    ----------
    capacity_id : uuid.UUID, default=None
        The capacity Id.
        Defaults to None which resolves to the capacity name of the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the capacity name of the workspace of the notebook.

    Returns
    -------
    str
        The capacity name.
    """
    from sempy_labs._capacities import list_capacities

    if capacity_id is None:
        return get_capacity_name()

    dfC = list_capacities()
    dfC_filt = dfC[dfC["Id"] == capacity_id]

    if dfC_filt.empty:
        raise ValueError(
            f"{icons.red_dot} The '{capacity_id}' capacity Id does not exist."
        )

    return dfC_filt["Display Name"].iloc[0]


def resolve_capacity_id(capacity: Optional[str | UUID] = None, **kwargs) -> UUID:
    """
    Obtains the capacity Id for a given capacity name.

    Parameters
    ----------
    capacity : str | uuid.UUID, default=None
        The capacity name or ID.
        Defaults to None which resolves to the capacity id of the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the capacity name of the workspace of the notebook.

    Returns
    -------
    uuid.UUID
        The capacity Id.
    """
    from sempy_labs._capacities import list_capacities

    if "capacity_name" in kwargs:
        capacity = kwargs["capacity_name"]
        print(
            f"{icons.warning} The 'capacity_name' parameter is deprecated. Please use 'capacity' instead."
        )

    if capacity is None:
        return get_capacity_id()
    if _is_valid_uuid(capacity):
        return capacity

    dfC = list_capacities()
    dfC_filt = dfC[dfC["Display Name"] == capacity]

    if dfC_filt.empty:
        raise ValueError(f"{icons.red_dot} The '{capacity}' capacity does not exist.")

    return dfC_filt["Id"].iloc[0]


def retry(sleep_time: int, timeout_error_message: str):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start_time = datetime.datetime.now()
            timeout = datetime.timedelta(minutes=1)
            while datetime.datetime.now() - start_time <= timeout:
                try:
                    return func(*args, **kwargs)
                except Exception:
                    time.sleep(sleep_time)
            raise TimeoutError(timeout_error_message)

        return wrapper

    return decorator


def lro(
    client,
    response,
    status_codes: Optional[List[str]] = [200, 202],
    sleep_time: Optional[int] = 1,
    return_status_code: bool = False,
):

    if response.status_code not in status_codes:
        raise FabricHTTPException(response)
    if response.status_code == status_codes[0]:
        if return_status_code:
            result = response.status_code
        else:
            result = response
    if response.status_code == status_codes[1]:
        operationId = response.headers["x-ms-operation-id"]
        response = client.get(f"/v1/operations/{operationId}")
        response_body = json.loads(response.content)
        while response_body["status"] not in ["Succeeded", "Failed"]:
            time.sleep(sleep_time)
            response = client.get(f"/v1/operations/{operationId}")
            response_body = json.loads(response.content)
        if response_body["status"] != "Succeeded":
            raise FabricHTTPException(response)
        if return_status_code:
            result = response.status_code
        else:
            response = client.get(f"/v1/operations/{operationId}/result")
            result = response

    return result


def pagination(client, response):

    responses = []
    response_json = response.json()
    responses.append(response_json)

    # Check for pagination
    continuation_token = response_json.get("continuationToken")
    continuation_uri = response_json.get("continuationUri")

    # Loop to handle pagination
    while continuation_token is not None:
        response = client.get(continuation_uri)
        response_json = response.json()
        responses.append(response_json)

        # Update the continuation token and URI for the next iteration
        continuation_token = response_json.get("continuationToken")
        continuation_uri = response_json.get("continuationUri")

    return responses


def resolve_deployment_pipeline_id(deployment_pipeline: str | UUID) -> UUID:
    """
    Obtains the Id for a given deployment pipeline.

    Parameters
    ----------
    deployment_pipeline : str | uuid.UUID
        The deployment pipeline name or ID.

    Returns
    -------
    uuid.UUID
        The deployment pipeline Id.
    """

    from sempy_labs._deployment_pipelines import list_deployment_pipelines

    if _is_valid_uuid(deployment_pipeline):
        return deployment_pipeline
    else:

        dfP = list_deployment_pipelines()
        dfP_filt = dfP[dfP["Deployment Pipeline Name"] == deployment_pipeline]
        if len(dfP_filt) == 0:
            raise ValueError(
                f"{icons.red_dot} The '{deployment_pipeline}' deployment pipeline is not valid."
            )
        return dfP_filt["Deployment Pipeline Id"].iloc[0]


class FabricTokenCredential(TokenCredential):

    def get_token(
        self,
        scopes: str,
        claims: Optional[str] = None,
        tenant_id: Optional[str] = None,
        enable_cae: bool = False,
        **kwargs: any,
    ) -> AccessToken:

        import notebookutils

        token = notebookutils.credentials.getToken("storage")
        return AccessToken(token, 0)


def _get_adls_client(account_name):

    from azure.storage.filedatalake import DataLakeServiceClient

    account_url = f"https://{account_name}.dfs.core.windows.net"

    return DataLakeServiceClient(account_url, credential=FabricTokenCredential())


def _get_blob_client(workspace_id: UUID, item_id: UUID):

    from azure.storage.blob import BlobServiceClient

    endpoint = _get_fabric_context_setting(name="trident.onelake.endpoint").replace(
        ".dfs.", ".blob."
    )
    url = f"https://{endpoint}/{workspace_id}/{item_id}"

    # account_url = f"https://{account_name}.blob.core.windows.net"

    return BlobServiceClient(url, credential=FabricTokenCredential())


def resolve_warehouse_id(
    warehouse: str | UUID, workspace: Optional[str | UUID]
) -> UUID:
    """
    Obtains the Id for a given warehouse.

    Parameters
    ----------
    warehouse : str | uuid.UUID
        The warehouse name or ID.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID in which the semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    uuid.UUID
        The warehouse Id.
    """

    return resolve_item_id(item=warehouse, type="Warehouse", workspace=workspace)


def get_language_codes(languages: str | List[str]):

    if isinstance(languages, str):
        languages = [languages]

    for i, lang in enumerate(languages):
        for k, v in icons.language_map.items():
            if v == lang.capitalize():
                languages[i] = k
                break

    return languages


def _get_azure_token_credentials(
    key_vault_uri: str,
    key_vault_tenant_id: str,
    key_vault_client_id: str,
    key_vault_client_secret: str,
    audience: str = "https://management.azure.com/.default",
) -> Tuple[str, str, dict]:

    import notebookutils
    from azure.identity import ClientSecretCredential

    # "https://analysis.windows.net/powerbi/api/.default"

    tenant_id = notebookutils.credentials.getSecret(key_vault_uri, key_vault_tenant_id)
    client_id = notebookutils.credentials.getSecret(key_vault_uri, key_vault_client_id)
    client_secret = notebookutils.credentials.getSecret(
        key_vault_uri, key_vault_client_secret
    )

    credential = ClientSecretCredential(
        tenant_id=tenant_id, client_id=client_id, client_secret=client_secret
    )

    token = credential.get_token(audience).token

    headers = {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json",
    }

    return token, credential, headers


def convert_to_alphanumeric_lowercase(input_string):

    # Removes non-alphanumeric characters
    cleaned_string = re.sub(r"[^a-zA-Z0-9]", "", input_string)
    cleaned_string = cleaned_string.lower()

    return cleaned_string


def resolve_environment_id(
    environment: str | UUID, workspace: Optional[str | UUID] = None
) -> UUID:
    """
    Obtains the environment Id for a given environment.

    Parameters
    ----------
    environment: str | uuid.UUID
        Name of the environment.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID in which the semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    uuid.UUID
        The environment Id.
    """

    return resolve_item_id(item=environment, type="Environment", workspace=workspace)


def _make_clickable(val):

    return f'<a target="_blank" href="{val}">{val}</a>'


def convert_to_friendly_case(text: str) -> str:
    """
    Converts a string of pascal/camel/snake case to business-friendly case.

    Parameters
    ----------
    text : str
        The text to convert.

    Returns
    -------
    str
        Text converted into a business-friendly text.
    """
    if text is not None:
        text = text.replace("_", " ")
        # Insert space before each capital letter, avoiding double spaces
        text = re.sub(r"(?<!\s)(?=[A-Z])", " ", text)
        # Strip leading/trailing whitespace and capitalize the first letter of each word
        text = text.strip().title()

    return text


def resolve_notebook_id(
    notebook: str | UUID, workspace: Optional[str | UUID] = None
) -> UUID:
    """
    Obtains the notebook Id for a given notebook.

    Parameters
    ----------
    notebook: str | uuid.UUID
        Name or ID of the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID in which the semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    uuid.UUID
        The notebook Id.
    """

    return resolve_item_id(item=notebook, type="Notebook", workspace=workspace)


def generate_guid():

    return str(uuid.uuid4())


def _get_column_aggregate(
    table_name: str,
    column_name: str = "RunId",
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
    function: str = "max",
    default_value: int = 0,
) -> int:

    from pyspark.sql.functions import approx_count_distinct
    from pyspark.sql import functions as F

    function = function.upper()
    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    lakehouse_id = resolve_lakehouse_id(lakehouse, workspace)
    path = create_abfss_path(lakehouse_id, workspace_id, table_name)

    spark = _create_spark_session()
    df = spark.read.format("delta").load(path)

    if function in {"COUNTDISTINCT", "DISTINCTCOUNT"}:
        result = df.select(F.count_distinct(F.col(column_name)))
    elif "APPROX" in function:
        result = df.select(approx_count_distinct(column_name))
    else:
        result = df.selectExpr(f"{function}({column_name})")

    return result.collect()[0][0] or default_value


def _make_list_unique(my_list):

    return list(set(my_list))


def _get_partition_map(
    dataset: str, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:

    partitions = fabric.evaluate_dax(
        dataset=dataset,
        workspace=workspace,
        dax_string="""
    select [ID] AS [PartitionID], [TableID], [Name] AS [PartitionName] from $system.tmschema_partitions
    """,
    )

    tables = fabric.evaluate_dax(
        dataset=dataset,
        workspace=workspace,
        dax_string="""
    select [ID] AS [TableID], [Name] AS [TableName] from $system.tmschema_tables
    """,
    )

    partition_map = pd.merge(partitions, tables, on="TableID", how="left")
    partition_map["PartitionID"] = partition_map["PartitionID"].astype(str)
    partition_counts = partition_map.groupby("TableID")["PartitionID"].transform(
        "count"
    )
    partition_map["Object Name"] = partition_map.apply(
        lambda row: (
            f"'{row['TableName']}'[{row['PartitionName']}]"
            if partition_counts[row.name] > 1
            else row["TableName"]
        ),
        axis=1,
    )
    return partition_map


def _show_chart(spec, title):

    h = f"""
    <!DOCTYPE html>
    <html>
        <head>
            <script src="https://cdn.jsdelivr.net/npm/vega@5"></script>
            <script src="https://cdn.jsdelivr.net/npm/vega-lite@5"></script>
            <script src="https://cdn.jsdelivr.net/npm/vega-embed@6"></script>
            <style>
                table, th, td {{
                border: 10px solid #e7e9eb;
                border-collapse: collapse;
                }}
            </style>
        </head>
        <body>
            <table>
                <tr>
                    <td style="text-align: center;">
                        <h1>{title}</h1>
                    </td>
                </tr>
                <tr>
                    <td>
                        <div id="vis"></div>
                    </td>
                </tr>
            </table>
            <script type="text/javascript">
                var spec = {spec};
                var opt = {{"renderer": "canvas", "actions": false}};
                vegaEmbed("#vis", spec, opt);
            </script>
        </body>
    </html>"""

    display(HTML(h))


def _process_and_display_chart(df, title, widget):

    # Convert time columns to milliseconds
    df["Start"] = df["Start Time"].astype(np.int64) / int(1e6)
    df["End"] = df["End Time"].astype(np.int64) / int(1e6)

    # Calculate the time offset for proper Gantt chart rendering
    Offset = min(df["Start"])
    df["Start"] = df["Start"] - Offset
    df["End"] = df["End"] - Offset

    # Vega-Lite spec for Gantt chart
    spec = (
        """{
        "$schema": "https://vega.github.io/schema/vega-lite/v5.json",
        "description": "A simple bar chart with ranged data (aka Gantt Chart).",
        "data": { "values": """
        + df.to_json(orient="records")
        + """ },
        "width": 700,
        "height": 400,
        "mark": "bar",
        "encoding": {
            "y": {
                "field": "Object Name",
                "type": "ordinal",
                "axis": {
                    "labelFontSize": 15,
                    "titleFontSize": 20,
                    "title": "Object"
                }
            },
            "x": {
                "field": "Start",
                "type": "quantitative",
                "title": "milliseconds",
                "axis": {
                    "titleFontSize": 20
                }
            },
            "x2": {"field": "End"},
            "color": {
                "field": "Event Subclass",
                "scale": {
                    "domain": ["Process", "ExecuteSql"],
                    "range": ["#FFC000","#0070C0"]
                },
                "legend": {
                    "labelFontSize": 20,
                    "titleFontSize": 20,
                    "title": "Event Type"
                }
            },
            "tooltip": [
                {"field": "Duration", "type": "quantitative", "format": ","},
                {"field": "Cpu Time", "type": "quantitative", "format": ","},
                {"field": "Event Subclass", "type": "nominal"}
            ]
        }
    }"""
    )

    with widget:
        widget.clear_output(wait=True)
    _show_chart(spec, title=title)


def _convert_data_type(input_data_type: str) -> str:

    data_type_mapping = {
        "string": "String",
        "int": "Int64",
        "tinyint": "Int64",
        "smallint": "Int64",
        "bigint": "Int64",
        "boolean": "Boolean",
        "timestamp": "DateTime",
        "date": "DateTime",
        "double": "Double",
        "float": "Double",
        "binary": "Boolean",
    }

    if "decimal" in input_data_type:
        return "Decimal"
    else:
        return data_type_mapping.get(input_data_type)


def _is_valid_uuid(
    guid: str,
):
    """
    Validates if a string is a valid GUID in version 4

    Parameters
    ----------
    guid : str
        GUID to be validated.

    Returns
    -------
    bool
        Boolean that indicates if the string is a GUID or not.
    """

    try:
        UUID(str(guid), version=4)
        return True
    except ValueError:
        return False


def _get_fabric_context_setting(name: str):

    from synapse.ml.internal_utils.session_utils import get_fabric_context

    return get_fabric_context().get(name)


def get_tenant_id():

    _get_fabric_context_setting(name="trident.tenant.id")


def _base_api(
    request: str,
    client: str = "fabric",
    method: str = "get",
    payload: Optional[str] = None,
    status_codes: Optional[int] = 200,
    uses_pagination: bool = False,
    lro_return_json: bool = False,
    lro_return_status_code: bool = False,
):

    from sempy_labs._authentication import _get_headers

    if (lro_return_json or lro_return_status_code) and status_codes is None:
        status_codes = [200, 202]

    if isinstance(status_codes, int):
        status_codes = [status_codes]

    if client == "fabric":
        c = fabric.FabricRestClient()
    elif client == "fabric_sp":
        c = fabric.FabricRestClient(token_provider=auth.token_provider.get())
    elif client in ["azure", "graph"]:
        pass
    else:
        raise ValueError(f"{icons.red_dot} The '{client}' client is not supported.")

    if client not in ["azure", "graph"]:
        if method == "get":
            response = c.get(request)
        elif method == "delete":
            response = c.delete(request)
        elif method == "post":
            response = c.post(request, json=payload)
        elif method == "patch":
            response = c.patch(request, json=payload)
        elif method == "put":
            response = c.put(request, json=payload)
        else:
            raise NotImplementedError
    else:
        headers = _get_headers(auth.token_provider.get(), audience=client)
        if client == "graph":
            url = f"https://graph.microsoft.com/v1.0/{request}"
        elif client == "azure":
            url = request
        else:
            raise NotImplementedError
        response = requests.request(
            method.upper(),
            url,
            headers=headers,
            json=payload,
        )

    if lro_return_json:
        return lro(c, response, status_codes).json()
    elif lro_return_status_code:
        return lro(c, response, status_codes, return_status_code=True)
    else:
        if response.status_code not in status_codes:
            raise FabricHTTPException(response)
        if uses_pagination:
            responses = pagination(c, response)
            return responses
        else:
            return response


def _create_dataframe(columns: dict) -> pd.DataFrame:

    return pd.DataFrame(columns=list(columns.keys()))


def _update_dataframe_datatypes(dataframe: pd.DataFrame, column_map: dict):
    """
    Updates the datatypes of columns in a pandas dataframe based on a column map.

    Example:
    {
        "Order": "int",
        "Public": "bool",
    }
    """

    for column, data_type in column_map.items():
        if column in dataframe.columns:
            if data_type == "int":
                dataframe[column] = dataframe[column].astype(int)
            elif data_type == "bool":
                dataframe[column] = dataframe[column].astype(bool)
            elif data_type == "float":
                dataframe[column] = dataframe[column].astype(float)
            elif data_type == "datetime":
                dataframe[column] = pd.to_datetime(dataframe[column])
            # This is for a special case in admin.list_reports where datetime itself does not work. Coerce fixes the issue.
            elif data_type == "datetime_coerce":
                dataframe[column] = pd.to_datetime(dataframe[column], errors="coerce")
            # This is for list_synonyms since the weight column is float and can have NaN values.
            elif data_type == "float_fillna":
                dataframe[column] = dataframe[column].fillna(0).astype(float)
            # This is to avoid NaN values in integer columns (for delta analyzer)
            elif data_type == "int_fillna":
                dataframe[column] = dataframe[column].fillna(0).astype(int)
            elif data_type in ["str", "string"]:
                dataframe[column] = dataframe[column].astype(str)
            else:
                raise NotImplementedError


def _print_success(item_name, item_type, workspace_name, action="created"):
    if action == "created":
        print(
            f"{icons.green_dot} The '{item_name}' {item_type} has been successfully created in the '{workspace_name}' workspace."
        )
    elif action == "deleted":
        print(
            f"{icons.green_dot} The '{item_name}' {item_type} has been successfully deleted from the '{workspace_name}' workspace."
        )
    else:
        raise NotImplementedError


def _pure_python_notebook() -> bool:

    from sempy.fabric._environment import _on_jupyter

    return _on_jupyter()


def _create_spark_session():

    if _pure_python_notebook():
        raise ValueError(
            f"{icons.red_dot} This function is only available in a PySpark notebook."
        )

    from pyspark.sql import SparkSession

    return SparkSession.builder.getOrCreate()


def _read_delta_table(path: str):

    spark = _create_spark_session()

    return spark.read.format("delta").load(path)


def _delta_table_row_count(table_name: str) -> int:

    spark = _create_spark_session()

    return spark.table(table_name).count()


def _run_spark_sql_query(query):

    spark = _create_spark_session()

    return spark.sql(query)


def _mount(
    lakehouse: Optional[str | UUID] = None, workspace: Optional[str | UUID] = None
) -> str:
    """
    Mounts a lakehouse to a notebook if it is not already mounted. Returns the local path to the lakehouse.
    """

    import notebookutils

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace=workspace)
    (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id(
        lakehouse=lakehouse, workspace=workspace
    )

    # Hide display mounts
    current_setting = ""
    try:
        current_setting = notebookutils.conf.get(
            "spark.notebookutils.displaymountpoint.enabled"
        )
        notebookutils.conf.set("spark.notebookutils.displaymountpoint.enabled", "false")
    except Exception:
        pass

    lake_path = create_abfss_path(lakehouse_id, workspace_id)
    mounts = notebookutils.fs.mounts()
    mount_point = f"/{workspace_name.replace(' ', '')}{lakehouse_name.replace(' ', '')}"
    if not any(i.get("source") == lake_path for i in mounts):
        # Mount lakehouse if not mounted
        notebookutils.fs.mount(lake_path, mount_point)
        print(
            f"{icons.green_dot} Mounted the '{lakehouse_name}' lakehouse within the '{workspace_name}' to the notebook."
        )

    mounts = notebookutils.fs.mounts()

    # Set display mounts to original setting
    try:
        if current_setting != "false":
            notebookutils.conf.set(
                "spark.notebookutils.displaymountpoint.enabled", "true"
            )
    except Exception:
        pass

    local_path = next(
        i.get("localPath") for i in mounts if i.get("source") == lake_path
    )

    return local_path


def _get_or_create_workspace(
    workspace: str,
    capacity: Optional[str | UUID] = None,
    description: Optional[str] = None,
) -> Tuple[str, UUID]:

    capacity_id = resolve_capacity_id(capacity)
    dfW = fabric.list_workspaces()
    dfW_filt_name = dfW[dfW["Name"] == workspace]
    dfW_filt_id = dfW[dfW["Id"] == workspace]

    # Workspace already exists
    if (not dfW_filt_name.empty) or (not dfW_filt_id.empty):
        print(f"{icons.green_dot} The '{workspace}' workspace already exists.")
        (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
        return (workspace_name, workspace_id)

    # Do not create workspace with name of an ID
    if _is_valid_uuid(workspace):
        raise ValueError(f"{icons.warning} Must enter a workspace name, not an ID.")

    print(f"{icons.in_progress} Creating the '{workspace}' workspace...")
    workspace_id = fabric.create_workspace(
        display_name=workspace, capacity_id=capacity_id, description=description
    )
    print(
        f"{icons.green_dot} The '{workspace}' workspace has been successfully created."
    )

    return (workspace, workspace_id)


def _get_or_create_lakehouse(
    lakehouse: str,
    workspace: Optional[str | UUID] = None,
    description: Optional[str] = None,
) -> Tuple[str, UUID]:

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    dfI = fabric.list_items(type="Lakehouse", workspace=workspace)
    dfI_filt_name = dfI[dfI["Display Name"] == lakehouse]
    dfI_filt_id = dfI[dfI["Id"] == lakehouse]

    if (not dfI_filt_name.empty) or (not dfI_filt_id.empty):
        print(f"{icons.green_dot} The '{lakehouse}' lakehouse already exists.")
        (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id(
            lakehouse=lakehouse, workspace=workspace
        )
        return (lakehouse_name, lakehouse_id)
    if _is_valid_uuid(lakehouse):
        raise ValueError(f"{icons.warning} Must enter a lakehouse name, not an ID.")

    print(f"{icons.in_progress} Creating the '{lakehouse}' lakehouse...")
    lakehouse_id = fabric.create_lakehouse(
        display_name=lakehouse, workspace=workspace, description=description
    )
    print(
        f"{icons.green_dot} The '{lakehouse}' lakehouse has been successfully created within the '{workspace_name}' workspace."
    )

    return (lakehouse, lakehouse_id)


def _get_or_create_warehouse(
    warehouse: str,
    workspace: Optional[str | UUID] = None,
    description: Optional[str] = None,
) -> Tuple[str, UUID]:

    from sempy_labs._warehouses import create_warehouse

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    dfI = fabric.list_items(type="Warehouse", workspace=workspace)
    dfI_filt_name = dfI[dfI["Display Name"] == warehouse]
    dfI_filt_id = dfI[dfI["Id"] == warehouse]

    if (not dfI_filt_name.empty) or (not dfI_filt_id.empty):
        print(f"{icons.green_dot} The '{warehouse}' warehouse already exists.")
        (warehouse_name, warehouse_id) = resolve_item_name_and_id(
            warehouse=warehouse, type="Warehouse", workspace=workspace
        )
        return (warehouse_name, warehouse_id)
    if _is_valid_uuid(warehouse):
        raise ValueError(f"{icons.warning} Must enter a warehouse name, not an ID.")

    print(f"{icons.in_progress} Creating the '{warehouse}' warehouse...")
    warehouse_id = create_warehouse(
        display_name=warehouse, workspace=workspace, description=description
    )
    print(
        f"{icons.green_dot} The '{warehouse}' warehouse has been successfully created within the '{workspace_name}' workspace."
    )

    return (warehouse, warehouse_id)


def _xml_to_dict(element):
    data = {element.tag: {} if element.attrib else None}
    children = list(element)
    if children:
        temp_dict = {}
        for child in children:
            child_dict = _xml_to_dict(child)
            for key, value in child_dict.items():
                if key in temp_dict:
                    if isinstance(temp_dict[key], list):
                        temp_dict[key].append(value)
                    else:
                        temp_dict[key] = [temp_dict[key], value]
                else:
                    temp_dict[key] = value
        data[element.tag] = temp_dict
    else:
        data[element.tag] = (
            element.text.strip() if element.text and element.text.strip() else None
        )
    return data



from sempy_labs.graph._groups import (
    list_groups,
    list_group_owners,
    list_group_members,
    add_group_members,
    add_group_owners,
    resolve_group_id,
    renew_group,
)
from sempy_labs.graph._users import (
    resolve_user_id,
    get_user,
    list_users,
    send_mail,
)
from sempy_labs.graph._teams import (
    list_teams,
)

__all__ = [
    "list_groups",
    "list_group_owners",
    "list_group_members",
    "add_group_members",
    "add_group_owners",
    "renew_group",
    "resolve_group_id",
    "resolve_user_id",
    "get_user",
    "list_users",
    "send_mail",
    "list_teams",
]



import pandas as pd
from sempy_labs._helper_functions import (
    _is_valid_uuid,
    _base_api,
    _update_dataframe_datatypes,
    _create_dataframe,
)
import sempy_labs._icons as icons
from uuid import UUID


def list_deployment_pipelines() -> pd.DataFrame:
    """
    Shows a list of deployment pipelines the user can access.

    This is a wrapper function for the following API: `Deployment Pipelines - List Deployment Pipelines <https://learn.microsoft.com/rest/api/fabric/core/deployment-pipelines/list-deployment-pipelines>`_.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of deployment pipelines the user can access.
    """

    columns = {
        "Deployment Pipeline Id": "string",
        "Deployment Pipeline Name": "string",
        "Description": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request="/v1/deploymentPipelines",
        status_codes=200,
        uses_pagination=True,
    )

    for r in responses:
        for v in r.get("value", []):
            new_data = {
                "Deployment Pipeline Id": v.get("id"),
                "Deployment Pipeline Name": v.get("displayName"),
                "Description": v.get("description"),
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def list_deployment_pipeline_stages(deployment_pipeline: str | UUID) -> pd.DataFrame:
    """
    Shows the specified deployment pipeline stages.

    This is a wrapper function for the following API: `Deployment Pipelines - List Deployment Pipeline Stages <https://learn.microsoft.com/rest/api/fabric/core/deployment-pipelines/list-deployment-pipeline-stages>`_.

    Parameters
    ----------
    deployment_pipeline : str | uuid.UUID
        The deployment pipeline name or ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the specified deployment pipeline stages.
    """

    from sempy_labs._helper_functions import resolve_deployment_pipeline_id

    columns = {
        "Deployment Pipeline Stage Id": "string",
        "Deployment Pipeline Stage Name": "string",
        "Order": "int",
        "Description": "string",
        "Workspace Id": "string",
        "Workspace Name": "string",
        "Public": "bool",
    }
    df = _create_dataframe(columns=columns)

    deployment_pipeline_id = resolve_deployment_pipeline_id(
        deployment_pipeline=deployment_pipeline
    )

    responses = _base_api(
        request=f"/v1/deploymentPipelines/{deployment_pipeline_id}/stages",
        status_codes=200,
        uses_pagination=True,
    )

    for r in responses:
        for v in r.get("value", []):
            new_data = {
                "Deployment Pipeline Stage Id": v.get("id"),
                "Deployment Pipeline Stage Name": v.get("displayName"),
                "Description": v.get("description"),
                "Order": v.get("order"),
                "Workspace Id": v.get("workspaceId"),
                "Workspace Name": v.get("workspaceName"),
                "Public": v.get("isPublic"),
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


def list_deployment_pipeline_stage_items(
    deployment_pipeline: str | UUID,
    stage: str | UUID,
) -> pd.DataFrame:
    """
    Shows the supported items from the workspace assigned to the specified stage of the specified deployment pipeline.

    This is a wrapper function for the following API: `Deployment Pipelines - List Deployment Pipeline Stage Items <https://learn.microsoft.com/rest/api/fabric/core/deployment-pipelines/list-deployment-pipeline-stage-items>`_.

    Parameters
    ----------
    deployment_pipeline : str | uuid.UUID
        The deployment pipeline name or ID.
    stage : str | uuid.UUID
        The deployment pipeline stage name or ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the supported items from the workspace assigned to the specified stage of the specified deployment pipeline.
    """

    from sempy_labs._helper_functions import resolve_deployment_pipeline_id

    columns = {
        "Deployment Pipeline Stage Item Id": "string",
        "Deployment Pipeline Stage Item Name": "string",
        "Item Type": "string",
        "Source Item Id": "string",
        "Target Item Id": "string",
        "Last Deployment Time": "string",
    }
    df = _create_dataframe(columns=columns)

    deployment_pipeline_id = resolve_deployment_pipeline_id(
        deployment_pipeline=deployment_pipeline
    )

    def resolve_deployment_pipeline_stage_id(
        deployment_pipeline_id: UUID, stage: str | UUID
    ):

        dfPS = list_deployment_pipeline_stages(
            deployment_pipeline=deployment_pipeline_id
        )

        if _is_valid_uuid(stage):
            dfPS_filt = dfPS[dfPS["Deployment Pipeline Stage Id"] == stage]
        else:
            dfPS_filt = dfPS[dfPS["Deployment Pipeline Stage Name"] == stage]
        if dfPS.empty:
            raise ValueError(
                f"{icons.red_dot} The '{stage}' stage does not exist within the '{deployment_pipeline}' deployment pipeline."
            )
        return dfPS_filt["Deployment Pipeline Stage Id"].iloc[0]

    stage_id = resolve_deployment_pipeline_stage_id(deployment_pipeline_id, stage)

    responses = _base_api(
        request=f"/v1/deploymentPipelines/{deployment_pipeline_id}/stages/{stage_id}/items",
        status_codes=200,
        uses_pagination=True,
    )

    for r in responses:
        for v in r.get("value", []):
            new_data = {
                "Deployment Pipeline Stage Item Id": v.get("itemId"),
                "Deployment Pipeline Stage Item Name": v.get("itemDisplayName"),
                "Item Type": v.get("itemType"),
                "Source Item Id": v.get("sourceItemId"),
                "Target Item Id": v.get("targetItemId"),
                "Last Deployment Time": v.get("lastDeploymentTime"),
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df



from sempy_labs.admin._users import (
    list_access_entities,
    list_user_subscriptions,
)
from sempy_labs.admin._workspaces import (
    add_user_to_workspace,
    delete_user_from_workspace,
    restore_deleted_workspace,
)
from sempy_labs.admin._artifacts import (
    list_unused_artifacts,
)
from sempy_labs.admin._shared import (
    list_widely_shared_artifacts,
)
from sempy_labs.admin._datasets import (
    list_datasets,
    list_dataset_users,
)
from sempy_labs.admin._apps import (
    list_apps,
    list_app_users,
)
from sempy_labs.admin._reports import (
    list_reports,
    list_report_users,
    list_report_subscriptions,
)
from sempy_labs.admin._activities import (
    list_activity_events,
)
from sempy_labs.admin._scanner import (
    scan_workspaces,
)
from sempy_labs.admin._capacities import (
    patch_capacity,
    list_capacities,
    get_capacity_assignment_status,
    get_capacity_state,
    list_capacity_users,
)
from sempy_labs.admin._tenant import (
    list_tenant_settings,
    delete_capacity_tenant_setting_override,
    update_tenant_setting,
    update_capacity_tenant_setting_override,
    list_workspaces_tenant_settings_overrides,
    list_capacity_tenant_settings_overrides,
    list_capacities_delegated_tenant_settings,
    list_domain_tenant_settings_overrides,
)
from sempy_labs.admin._basic_functions import (
    assign_workspaces_to_capacity,
    unassign_workspaces_from_capacity,
    list_workspaces,
    list_workspace_access_details,
    list_modified_workspaces,
    list_workspace_users,
)
from sempy_labs.admin._domains import (
    list_domains,
    list_domain_workspaces,
    assign_domain_workspaces,
    assign_domain_workspaces_by_capacities,
    create_domain,
    update_domain,
    delete_domain,
    resolve_domain_id,
    unassign_domain_workspaces,
    unassign_all_domain_workspaces,
)
from sempy_labs.admin._items import (
    list_item_access_details,
    list_items,
)
from sempy_labs.admin._external_data_share import (
    list_external_data_shares,
    revoke_external_data_share,
)
from sempy_labs.admin._git import (
    list_git_connections,
)

__all__ = [
    "list_items",
    "list_workspace_access_details",
    "list_access_entities",
    "list_item_access_details",
    "list_datasets",
    "list_workspaces",
    "assign_workspaces_to_capacity",
    "list_capacities",
    "list_tenant_settings",
    "list_domains",
    "list_domain_workspaces",
    "assign_domain_workspaces",
    "assign_domain_workspaces_by_capacities",
    "create_domain",
    "update_domain",
    "delete_domain",
    "resolve_domain_id",
    "unassign_domain_workspaces",
    "unassign_all_domain_workspaces",
    "list_capacities_delegated_tenant_settings",
    "unassign_workspaces_from_capacity",
    "list_external_data_shares",
    "revoke_external_data_share",
    "list_activity_events",
    "list_modified_workspaces",
    "list_git_connections",
    "list_reports",
    "get_capacity_assignment_status",
    "scan_workspaces",
    "get_capacity_state",
    "list_apps",
    "list_app_users",
    "list_dataset_users",
    "list_report_users",
    "patch_capacity",
    "list_workspace_users",
    "list_widely_shared_artifacts",
    "delete_capacity_tenant_setting_override",
    "update_tenant_setting",
    "update_capacity_tenant_setting_override",
    "list_workspaces_tenant_settings_overrides",
    "list_capacity_tenant_settings_overrides",
    "list_capacities_delegated_tenant_settings",
    "list_domain_tenant_settings_overrides",
    "list_unused_artifacts",
    "add_user_to_workspace",
    "delete_user_from_workspace",
    "restore_deleted_workspace",
    "list_capacity_users",
    "list_user_subscriptions",
    "list_report_subscriptions",
]
